# Copyrigh 2016 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import collections
import copy

import mock
import sys

import base64
import netaddr
from oslo_serialization import jsonutils

from neutron.tests import base as base_test

from networking_cisco.apps.saf.server.services.firewall.native import (
    fabric_setup_base as FP)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    asa_rest as asa)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    base)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    phy_asa)


TENANT_NAME = 'TenantA'
TENANT_ID = '0000-1111-2222-5555'
FW_ID = '0000-aaaa-bbbb-ccce'
FW_NAME = 'FwA'
POLCY_ID = '0000-aaaa-bbbb-cccc'
FW_TYPE = 'TE'
ROUTER_ID = '0000-aaaa-bbbb-5555'
RULE_ID = '0000-aaaa-bbbb-cccd'
PROTOCOL = 'tcp'
MGMT_IP = '3.3.3.3'
SRC_IP = '1.1.1.1'
DST_IP = '2.2.2.2'
SRC_PORT = 34
DST_PORT = 43
RULE_NAME = 'RuleA'
IN_SUBNET = '100.100.11.0'
IN_MASK = '255.255.255.0'
IN_SUBNET_AND_MASK = '100.100.11.0/24'
IN_START = '100.100.11.3'
IN_SEC_GW = '100.100.11.254'
IN_GW = '100.100.11.2'
IN_FABRIC_GW = '100.100.11.1'
IN_END = '100.100.11.254'
SEGMENTATION_ID = 87999
OUT_SEGMENTATION_ID = 88000
VLAN_ID = 770
OUT_VLAN_ID = 771
OUT_SUBNET = '200.200.11.0'
OUT_MASK = '255.255.255.0'
OUT_SUBNET_AND_MASK = '200.200.11.0/24'
RTR_SUBNET_AND_MASK = '9.9.9.0/24'
OUT_START = '200.200.11.3'
OUT_SEC_GW = '200.200.11.254'
OUT_GW = '200.200.11.2'
OUT_FABRIC_GW = '200.200.11.1'
OUT_END = '200.200.11.254'
INTF_IN = 'e1/1'
INTF_OUT = 'e1/2'

try:
    OrderedDict = collections.OrderedDict
except AttributeError:
    import ordereddict
    OrderedDict = ordereddict.OrderedDict


class FakeClass(object):
    """Fake class"""
    @classmethod
    def imitate(cls, *others):
        for other in others:
            for name in other.__dict__:
                try:
                    setattr(cls, name, mock.Mock())
                except (TypeError, AttributeError):
                    pass
        return cls

    @classmethod
    def set_return(cls, class_name, fn_name, return_val):
        getattr(cls, fn_name).return_value = return_val


class PhyAsaTest(base_test.BaseTestCase):
    """A test suite to exercise the Phy ASA Driver.  """

    def setUp(self):
        '''Setup for the test scripts '''
        super(PhyAsaTest, self).setUp()
        self._init_values()
        self.cfg_dict = self._fill_cfg()

        phy_asa.PhyAsa.__bases__ = (FakeClass.imitate(base.BaseDriver,
                                                      FP.FabricApi),)
        self.phy_asa = phy_asa.PhyAsa()
        self.phy_asa.initialize(self.cfg_dict)
        self.asa_rest = asa.Asa5585(MGMT_IP, 'user', 'user')
        self.headers = {'Content-Type': 'application/json'}
        self.byte_str = ('%s:%s' % ('user', 'user')).encode()
        self.base64string = base64.encodestring(
            self.byte_str).decode().replace('\n', '')
        if sys.version_info[0] < 3:
            self.url_open = 'urllib2.urlopen'
            self.url_req = 'urllib2.Request'
        else:
            self.url_open = 'urllib.request.urlopen'
            self.url_req = 'urllib.request.Request'

    def _init_values(self):
        self.tenant_name = TENANT_NAME
        self.tenant_id = TENANT_ID
        self.fw_id = FW_ID
        self.fw_name = FW_NAME
        self.rule_id = RULE_ID
        self.rule_dict = {'protocol': PROTOCOL, 'name': RULE_NAME,
                          'enabled': True, 'source_ip_address': SRC_IP,
                          'destination_ip_address': DST_IP,
                          'source_port': str(SRC_PORT),
                          'destination_port': str(DST_PORT),
                          'action': 'allow'}
        self.in_subnet = IN_SUBNET
        self.in_subnet_and_mask = IN_SUBNET_AND_MASK
        self.in_gw = IN_GW
        self.in_start = IN_START
        self.in_sec_gw = IN_SEC_GW
        self.in_end = IN_END
        self.out_subnet = OUT_SUBNET
        self.out_subnet_and_mask = OUT_SUBNET_AND_MASK
        self.out_gw = OUT_GW
        self.out_sec_gw = OUT_SEC_GW
        self.out_start = OUT_START
        self.out_end = OUT_END
        self.in_fabric_gw = IN_FABRIC_GW
        self.out_fabric_gw = OUT_FABRIC_GW
        self.segmentation_id = SEGMENTATION_ID
        self.out_segmentation_id = OUT_SEGMENTATION_ID
        self.vlan_id = VLAN_ID
        self.out_vlan_id = OUT_VLAN_ID
        fw_data_rule_dict = {self.rule_id: self.rule_dict}
        self.fw_data = {'tenant_name': self.tenant_name, 'fw_id': self.fw_id,
                        'fw_name': self.fw_name, 'rules': fw_data_rule_dict}

    def _fill_cfg(self):
        config = {'mgmt_ip_addr': MGMT_IP, 'user': 'user',
                  'pwd': 'user', 'interface_in': INTF_IN,
                  'interface_out': INTF_OUT}
        return config

    def test_phy_asa_init(self):
        '''Wrapper for the init'''
        pass

    def _get_modified_fw_data(self):
        new_fw_data = copy.deepcopy(self.fw_data)
        new_fw_data.get('rules').get(self.rule_id)['action'] = 'deny'
        return new_fw_data

    def _get_asa_command(self):
        cmds = ["conf t", "changeto system"]
        inside_int = INTF_IN + '.' + str(self.vlan_id)
        cmds.append("int " + inside_int)
        cmds.append("vlan " + str(self.vlan_id))
        outside_int = INTF_OUT + '.' + str(self.out_vlan_id)
        cmds.append("int " + outside_int)
        cmds.append("vlan " + str(self.out_vlan_id))
        cmds.append("context " + self.tenant_name)
        cmds.append("allocate-interface " + inside_int)
        cmds.append("allocate-interface " + outside_int)
        cmds.append("config-url disk0:/" + self.tenant_name + ".cfg")
        cmds.append("write memory")
        cmds.append("changeto context " + self.tenant_name)
        cmds.append("int " + inside_int)
        cmds.append("nameif Inside")
        cmds.append("security-level 100")
        cmds.append("ip address " + self.in_gw + " " + IN_MASK)
        cmds.append("int " + outside_int)
        cmds.append("nameif Outside")
        cmds.append("security-level 0")
        cmds.append("ip address " + self.out_gw + " " + OUT_MASK)

        cmds.append("router ospf 1")
        cmds.append("network " + self.in_gw + " " + IN_MASK + " area 0")
        cmds.append("network " + self.out_gw + " " + OUT_MASK + " area 0")
        cmds.append("area 0")
        cmds.append("route Outside 0.0.0.0 0.0.0.0 " + self.out_fabric_gw +
                    " 1")
        cmds.append("route Outside 0.0.0.0 0.0.0.0 " + self.out_sec_gw + " 1")
        cmds.append("end")
        cmds.append("write memory")

        data = {"commands": cmds}
        return data

    def _get_asa_del_command(self):
        cmds = ["conf t", "changeto system"]
        cmds.append("no context " + self.tenant_name + " noconfirm")
        inside_int = INTF_IN + '.' + str(self.vlan_id)
        outside_int = INTF_OUT + '.' + str(self.out_vlan_id)
        cmds.append("no interface " + inside_int)
        cmds.append("no interface " + outside_int)
        cmds.append("write memory")
        cmds.append("del /noconfirm disk0:/" + self.tenant_name + ".cfg")

        data = {"commands": cmds}
        return data

    def _get_asa_pol_command(self):
        cmds = ["conf t", "changeto context " + self.tenant_name]
        acl = "access-list "
        acl = (acl + self.tenant_name + " extended " +
               'permit' + " " +
               self.rule_dict['protocol'] + " " + SRC_IP + " " +
               '255.255.255.255' + " eq " + str(SRC_PORT) + " " +
               DST_IP + " " + '255.255.255.255' + " eq " + str(DST_PORT) + " ")
        cmds.append(acl)
        cmds.append("access-group " + self.tenant_name + " global")
        cmds.append("write memory")

        data = {"commands": cmds}
        return data

    def _get_asa_modf_pol_command(self):
        cmds = ["conf t", "changeto context " + self.tenant_name]
        acl = "access-list "
        acl = (acl + self.tenant_name + " extended " +
               'permit' + " " +
               self.rule_dict['protocol'] + " " + SRC_IP + " " +
               '255.255.255.255' + " eq " + str(SRC_PORT) + " " +
               DST_IP + " " + '255.255.255.255' + " eq " + str(DST_PORT) + " ")
        cmds.append("no " + acl)
        acl = ("access-list " + self.tenant_name + " extended " +
               'deny' + " " +
               self.rule_dict['protocol'] + " " + SRC_IP + " " +
               '255.255.255.255' + " eq " + str(SRC_PORT) + " " +
               DST_IP + " " + '255.255.255.255' + " eq " + str(DST_PORT) + " ")
        cmds.append(acl)
        cmds.append("access-group " + self.tenant_name + " global")
        cmds.append("write memory")

        data = {"commands": cmds}
        return data

    def test_build_acl_valid_ip(self):
        acl = self.asa_rest.build_acl_ip(netaddr.IPNetwork(self.in_fabric_gw))
        self.assertEqual(self.in_fabric_gw + ' 255.255.255.255 ', acl)

    def test_get_ip_address_valid(self):
        ip_address = self.asa_rest.get_ip_address(self.out_fabric_gw)
        self.assertEqual(netaddr.IPNetwork(self.out_fabric_gw), ip_address)

    def test_get_ip_address_null(self):
        ip_address = self.asa_rest.get_ip_address(None)
        self.assertEqual(netaddr.IPNetwork('0.0.0.0/0'), ip_address)

    def test_build_acl_ip_none(self):
        acl = self.asa_rest.build_acl_ip(netaddr.IPNetwork('0.0.0.0/0'))
        self.assertEqual("any ", acl)

    def test_build_acl_port_src(self):
        acl = self.asa_rest.build_acl_port(str(SRC_PORT), enabled=True)
        self.assertEqual('eq ' + str(SRC_PORT) + ' ', acl)

    def test_build_acl_port_dst(self):
        acl = self.asa_rest.build_acl_port(str(DST_PORT), enabled=True)
        self.assertEqual('eq ' + str(DST_PORT) + ' ', acl)

    def test_build_acl_port_not_enabled_dst(self):
        acl = self.asa_rest.build_acl_port(str(DST_PORT), enabled=False)
        self.assertEqual('eq ' + str(DST_PORT) + ' inactive', acl)

    def test_build_acl_port_range_dst(self):
        port_range = str(DST_PORT) + ' : ' + str(DST_PORT + 10)
        acl = self.asa_rest.build_acl_port(port_range, enabled=True)
        self.assertEqual('range ' + port_range.replace(':', ' ') + ' ', acl)

    def _test_create_fw(self):
        url = "https://" + MGMT_IP + "/api/cli"
        asa_payload = self._get_asa_command()
        asa_pol_payload = self._get_asa_pol_command()
        with mock.patch(self.url_req) as url_req,\
                mock.patch(self.url_open) as url_open:
            url_open.return_value.getcode.return_value = 202
            FakeClass.set_return(FP.FabricApi, 'get_in_ip_addr',
                                 {'subnet': self.in_subnet,
                                  'start': self.in_start,
                                  'sec_gateway': self.in_sec_gw,
                                  'gateway': self.in_fabric_gw,
                                  'end': self.in_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_srvc_node_ip_addr',
                                 self.in_gw)
            FakeClass.set_return(FP.FabricApi, 'get_out_srvc_node_ip_addr',
                                 self.out_gw)
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            parent = mock.MagicMock()
            parent.attach_mock(url_req, 'url_req')
            self.phy_asa.create_fw(self.tenant_id, self.fw_data)
        expected_calls = [
            mock.call.url_req(url, jsonutils.dumps(asa_payload), self.headers),
            mock.call.url_req().add_header("Authorization",
                                           "Basic %s" % self.base64string),
            mock.call.url_req(url, jsonutils.dumps(asa_pol_payload),
                              self.headers),
            mock.call.url_req().add_header("Authorization",
                                           "Basic %s" % self.base64string)]
        parent.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual(url_open.called, True)

    def test_create_fw(self):
        """Create FW Test """
        self._test_create_fw()

    def _test_delete_fw(self):
        url = "https://" + MGMT_IP + "/api/cli"
        asa_del_payload = self._get_asa_del_command()
        with mock.patch(self.url_req) as url_req,\
                mock.patch(self.url_open) as url_open:
            url_open.return_value.getcode.return_value = 202
            FakeClass.set_return(FP.FabricApi, 'get_in_ip_addr',
                                 {'subnet': self.in_subnet,
                                  'start': self.in_start,
                                  'sec_gateway': self.in_sec_gw,
                                  'gateway': self.in_fabric_gw,
                                  'end': self.in_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_srvc_node_ip_addr',
                                 self.in_gw)
            FakeClass.set_return(FP.FabricApi, 'get_out_srvc_node_ip_addr',
                                 self.out_gw)
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            parent = mock.MagicMock()
            parent.attach_mock(url_req, 'url_req')
            self.phy_asa.delete_fw(self.tenant_id, self.fw_data)
        expected_calls = [
            mock.call.url_req(url, jsonutils.dumps(asa_del_payload),
                              self.headers),
            mock.call.url_req().add_header("Authorization",
                                           "Basic %s" % self.base64string)]
        parent.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual(url_open.called, True)

    def test_delete_fw(self):
        """Delete FW Test """
        self._test_delete_fw()

    def _test_modify_fw(self):
        url = "https://" + MGMT_IP + "/api/cli"
        asa_payload = self._get_asa_command()
        asa_pol_payload = self._get_asa_pol_command()
        asa_modf_pol_payload = self._get_asa_modf_pol_command()
        with mock.patch(self.url_req) as url_req,\
                mock.patch(self.url_open) as url_open:
            url_open.return_value.getcode.return_value = 202
            FakeClass.set_return(FP.FabricApi, 'get_in_ip_addr',
                                 {'subnet': self.in_subnet,
                                  'start': self.in_start,
                                  'sec_gateway': self.in_sec_gw,
                                  'gateway': self.in_fabric_gw,
                                  'end': self.in_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_srvc_node_ip_addr',
                                 self.in_gw)
            FakeClass.set_return(FP.FabricApi, 'get_out_srvc_node_ip_addr',
                                 self.out_gw)
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            parent = mock.MagicMock()
            parent.attach_mock(url_req, 'url_req')
            self.phy_asa.create_fw(self.tenant_id, self.fw_data)
            new_fw_data = self._get_modified_fw_data()
            self.phy_asa.modify_fw(self.tenant_id, new_fw_data)
        expected_calls = [
            mock.call.url_req(url, jsonutils.dumps(asa_payload), self.headers),
            mock.call.url_req().add_header("Authorization",
                                           "Basic %s" % self.base64string),
            mock.call.url_req(url, jsonutils.dumps(asa_pol_payload),
                              self.headers),
            mock.call.url_req().add_header("Authorization",
                                           "Basic %s" % self.base64string),
            mock.call.url_req(url, jsonutils.dumps(asa_modf_pol_payload),
                              self.headers),
            mock.call.url_req().add_header("Authorization",
                                           "Basic %s" % self.base64string)]
        parent.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual(url_open.called, True)

    def test_modify_fw(self):
        """Modify FW Test """
        self._test_modify_fw()



# Copyright (c) 2016 Cisco Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_config import types


class RemainderOpt(cfg.DictOpt):

    """A greedy dictionary option

    This option will greedily parse any unparsed options in the config section
    this option is included in storing them in a dictionary structure. For
    example:

    A config file like this::

        [section_a]
        host1=value1
        host2=value2

    with a RemainderOpt option defined like::

        opts = [
            RemainderOpt('hosts')
        ]
        cfg.CONF.register_opts(opts, "section_a")

    Results in a dictionary like::

        hosts: {
           host1: value1,
           host2: value2
        }

    which is accessed using the oslo config object being used, for example::

        cfg.CONF['hosts']['host1']
    """

    def __init__(self, name, item_type=None, **kwargs):
        super(RemainderOpt, self).__init__(name, **kwargs)
        if not item_type:
            item_type = types.String()
        self.item_type = item_type

    def _get_opts_including_deprecated(self, namespace, group_name):
        opts = namespace._conf._get_group(group_name)._opts
        opt_names = list(opts)
        for name, opt in opts.items():
            opt_names.extend([x.name for x in opt['opt'].deprecated_opts])
        return opt_names

    def _get_from_namespace(self, namespace, group_name):
        opt_names = self._get_opts_including_deprecated(namespace,
                                                        group_name)
        result = {}
        for section in namespace._parsed:
            gk = section.get(group_name)
            if not gk:
                continue
            for key in gk:
                if key not in opt_names:
                    names = [(group_name, key)]
                    value = namespace._get_value(
                        names, positional=self.positional)
                    result[key] = self.item_type(value)
        return result


class SubsectionOpt(cfg.DictOpt):

    """An option for parsing multiple sections with sub-IDs

    This option allows to parse multiple defintions of the same config section
    which have unique IDs. For example::

        [switch:switch1]
        address=1.1.1.1
        password=p1

        [switch:switch2]
        address=2.2.2.2
        password=p2

    Both sections are type "switch" and will provide the same configuration
    options, but one is switch1 and the other is for switch2.

    Using this option this can be represented like::

        SubsectionOpt("switch",
                      dest="switches",
                      subopts=[StrOpt('address')
                               StrOpt('password')])

    When parsed the above config file example will result in a dictionary in
    the form::

        switches: {
            switch1: {
                address: 1.1.1.1,
                password: p1
            },
            switch2: {
                address: 2.2.2.2,
                password: p2
            }
        }
    """

    def __init__(self, name, subopts=None, **kwargs):
        super(SubsectionOpt, self).__init__(name, **kwargs)
        self.subopts = subopts or []

    def _get_from_namespace(self, namespace, group_name):
        identities = {}
        sections = cfg.CONF.list_all_sections()
        for section in sections:
            subsection, sep, ident = section.partition(':')
            if subsection.lower() != self.name.lower():
                continue
            cfg.CONF.register_opts(self.subopts, group=section)
            identities[ident] = cfg.CONF.get(section)
        return identities



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.tests import base

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cfg_agent import cfg_agent
from networking_cisco.plugins.cisco.cfg_agent.service_helpers import (
    routing_svc_helper as svc_helper)
from networking_cisco.plugins.cisco.cfg_agent.service_helpers import (
    routing_svc_helper_aci as aci_svc_helper)
from networking_cisco.tests.unit.cisco.cfg_agent import (
    test_routing_svc_helper as helper)
from networking_cisco.tests.unit.cisco.cfg_agent import cfg_agent_test_support

_uuid = uuidutils.generate_uuid

TEST_PHYS_IF = 'GigabitEthernet0/1/0'
TEST_VLAN = '3333'
TEST_GW_IP = '4.3.2.1'
TEST_CIDR = '4.3.2.0/24'
TEST_NET1 = 'mynewnet'
TEST_VLAN2 = '4444'
TEST_GW_IP2 = '5.4.3.1'
TEST_CIDR2 = '5.4.3.0/24'
TEST_NET2 = 'myothernet'


def create_hosting_info(vrf=None, net_name=TEST_NET1,
                        vlan=TEST_VLAN, gw_ip=TEST_GW_IP,
                        cidr=TEST_CIDR, if_config=None,
                        global_config=None, snat_subnets=None):
    if vrf is None:
        vrf = _uuid()
    hosting_info = {
        'vrf_id': vrf,
        'physical_interface': TEST_PHYS_IF,
        'network_name': net_name,
        'segmentation_id': vlan,
        'segmentation_id': vlan,
        'gateway_ip': gw_ip,
        'cidr_exposed': cidr,
    }
    if if_config:
        hosting_info['interface_config'] = if_config
    if global_config:
        hosting_info['global_config'] = global_config
    if snat_subnets:
        hosting_info['snat_subnets'] = snat_subnets
    return hosting_info


class TestBasicRoutingOperationsAci(helper.TestBasicRoutingOperations):

    def setUp(self):
        super(TestBasicRoutingOperationsAci, self).setUp()
        self.routing_helper = aci_svc_helper.RoutingServiceHelperAci(
            helper.HOST, self.conf, self.agent)
        self.routing_helper._internal_network_added = mock.Mock()
        self.routing_helper._external_gateway_added = mock.Mock()
        self.routing_helper._internal_network_removed = mock.Mock()
        self.routing_helper._external_gateway_removed = mock.Mock()
        self.routing_helper._enable_router_interface = mock.Mock()
        self.routing_helper._disable_router_interface = mock.Mock()
        self.driver = self._mock_driver_and_hosting_device(
            self.routing_helper)

    def test_process_router_throw_multiple_ipv4_subnets_error(self):
        ri, router = (
            self._test_process_router_throw_multiple_ipv4_subnets_error())
        router['gw_port'] = {'id': ''}
        self.assertRaises(svc_helper.MultipleIPv4SubnetsException,
                          self.routing_helper._process_router, ri)

    def test_process_router(self):
        super(TestBasicRoutingOperationsAci,
              self).test_process_router(test_admin_state=False)

    def test_process_msn_router(self):
        self._test_process_msn_router(test_admin_state=False)

    def test_process_router_2_rids_1_vrf(self):
        driver = self._mock_driver_and_hosting_device(self.routing_helper)
        router1, ports = self.prepare_router_data()
        ri1 = svc_helper.RouterInfo(router1['id'], router=router1)

        # Router #2 is like #1, except with some different IDs
        router2 = copy.deepcopy(router1)
        router2['id'] = _uuid()
        ri2 = svc_helper.RouterInfo(router2['id'], router=router2)
        h_info1 = create_hosting_info()
        h_info2 = copy.deepcopy(h_info1)
        ri1.router['hosting_info'] = h_info1
        ri2.router['hosting_info'] = h_info2

        driver._get_vrf_name = mock.Mock(
            return_value=ri1.router['hosting_info']['vrf_id'])
        self.routing_helper._process_router(ri1)

        vrf = ri1.router['hosting_info']['vrf_id']
        driver._get_vrf_name.assert_called_with(ri1)
        driver._do_create_vrf.assert_called_with(vrf)
        self.assertEqual(1, len(self.routing_helper._router_ids_by_vrf))
        self.assertEqual(1, len(self.routing_helper._router_ids_by_vrf[vrf]))

        driver._get_vrf_name.reset_mock()
        driver._do_create_vrf.reset_mock()

        self.routing_helper._process_router(ri2)

        driver._get_vrf_name.assert_called_with(ri2)
        driver._do_create_vrf.assert_not_called()
        self.assertEqual(1, len(self.routing_helper._router_ids_by_vrf))
        self.assertEqual(2, len(self.routing_helper._router_ids_by_vrf[vrf]))

        del ri1.router['gw_port']
        driver._get_vrf_name.reset_mock()

        self.routing_helper._process_router(ri1)
        driver._get_vrf_name.assert_called_with(ri1)
        driver._remove_vrf.assert_not_called()
        self.assertEqual(1, len(self.routing_helper._router_ids_by_vrf))
        self.assertEqual(1, len(self.routing_helper._router_ids_by_vrf[vrf]))

        del ri2.router['gw_port']
        driver._get_vrf_name.reset_mock()
        driver._remove_vrf.reset_mock()
        self.routing_helper._process_router(ri2)
        driver._get_vrf_name.assert_called_with(ri2)
        driver._remove_vrf.assert_called_with(ri2)
        self.assertEqual({}, self.routing_helper._router_ids_by_vrf)


def _mock_driver_and_hosting_device(svc_helper):
    svc_helper._dev_status.is_hosting_device_reachable = mock.MagicMock(
        return_value=True)
    driver = mock.MagicMock()
    svc_helper._drivermgr.get_driver = mock.Mock(return_value=driver)
    svc_helper._drivermgr.set_driver = mock.Mock(return_value=driver)
    return driver


class TestNetworkRoutingOperationsAci(
        base.BaseTestCase, cfg_agent_test_support.CfgAgentTestSupportMixin):

    def setUp(self):
        super(TestNetworkRoutingOperationsAci, self).setUp()
        self.agent = mock.Mock()
        self.conf = cfg.ConfigOpts()
        self.conf.register_opts(bc.core_opts)
        self.conf.register_opts(cfg_agent.OPTS, "cfg_agent")
        self.l3pluginApi_cls_p = mock.patch(
            'networking_cisco.plugins.cisco.cfg_agent.service_helpers.'
            'routing_svc_helper.CiscoRoutingPluginApi')
        l3plugin_api_cls = self.l3pluginApi_cls_p.start()
        self.plugin_api = mock.Mock()
        l3plugin_api_cls.return_value = self.plugin_api
        self.plugin_api.get_routers = mock.MagicMock()
        self.looping_call_p = mock.patch(
            'oslo_service.loopingcall.FixedIntervalLoopingCall')
        self.looping_call_p.start()
        mock.patch('neutron.common.rpc.create_connection').start()
        self.routing_helper = aci_svc_helper.RoutingServiceHelperAci(
            helper.HOST, self.conf, self.agent)
        self.routing_helper._external_gateway_added = mock.Mock()
        self.routing_helper._external_gateway_removed = mock.Mock()

    def _set_driver_port_mocks(self, driver):
        driver.internal_network_added = mock.Mock()
        driver.internal_network_removed = mock.Mock()
        driver.enable_internal_network_NAT = mock.Mock()
        driver.disable_internal_network_NAT = mock.Mock()

    def test_process_router_2_rids_1_vrf_1_network(self):
        driver = _mock_driver_and_hosting_device(self.routing_helper)
        self._set_driver_port_mocks(driver)

        router1, ports = self.prepare_router_data()
        ri1 = svc_helper.RouterInfo(router1['id'], router=router1)

        # Router #2 is like #1, except with some different IDs
        router2 = copy.deepcopy(router1)
        router2['id'] = _uuid()
        ri2 = svc_helper.RouterInfo(router2['id'], router=router2)
        h_info1 = create_hosting_info()
        h_info2 = copy.deepcopy(h_info1)
        ri1.router['hosting_info'] = h_info1
        ri2.router['hosting_info'] = h_info2

        ex_gw_port1 = ri1.router.get('gw_port')
        ex_gw_port2 = ri2.router.get('gw_port')
        ex_gw_port1['hosting_info'] = h_info1
        ex_gw_port2['hosting_info'] = h_info2
        vrf = ri1.router['hosting_info']['vrf_id']
        driver._get_vrf_name = mock.Mock(return_value=vrf)
        self.routing_helper._process_router(ri1)

        # remove change_details as they are recursive and prevent comparison
        del ports[0]['change_details']
        driver.internal_network_added.assert_called_with(
            ri1, ports[0])
        driver.enable_internal_network_NAT.assert_called_with(
            ri1, ports[0], ex_gw_port1)

        v_n_r_dict = self.routing_helper._router_ids_by_vrf_and_ext_net
        network_name = h_info1['network_name']
        self.assertEqual(1, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf]))
        self.assertEqual(1, len(v_n_r_dict[vrf][network_name]))

        driver.internal_network_added.reset_mock()
        driver.enable_internal_network_NAT.reset_mock()

        self.routing_helper._process_router(ri2)
        # remove change_details as they are recursive and prevent comparison
        del ri2.internal_ports[0]['change_details']
        driver.internal_network_added.assert_called_with(
            ri2, ports[0])
        driver.enable_internal_network_NAT.assert_called_with(
            ri2, ports[0], ex_gw_port2)
        network_name = h_info1['network_name']
        self.assertEqual(1, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf]))
        self.assertEqual(2, len(v_n_r_dict[vrf][network_name]))

        del ri1.router[bc.constants.INTERFACE_KEY]
        self.routing_helper._process_router(ri1)

        driver.internal_network_removed.assert_called_with(
            ri1, ports[0], itfc_deleted=False)
        driver.disable_internal_network_NAT.assert_called_with(
            ri1, ports[0], ex_gw_port1, itfc_deleted=False)
        self.assertEqual(1, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf]))
        self.assertEqual(1, len(v_n_r_dict[vrf][network_name]))

        driver.internal_network_removed.reset_mock()
        driver.disable_internal_network_NAT.reset_mock()

        del ri2.router[bc.constants.INTERFACE_KEY]
        self.routing_helper._process_router(ri2)

        # remove change_details as they are recursive and prevent comparison
        del ports[0]['change_details']
        del driver.internal_network_removed.call_args[0][1]['change_details']
        driver.internal_network_removed.assert_called_with(
            ri2, ports[0], itfc_deleted=True)
        driver.disable_internal_network_NAT.assert_called_with(
            ri2, ports[0], ex_gw_port2, itfc_deleted=True)
        self.assertEqual({}, v_n_r_dict)

    def test_process_router_2_rids_2_vrfs_1_network(self):
        driver = _mock_driver_and_hosting_device(self.routing_helper)
        self._set_driver_port_mocks(driver)

        router1, ports = self.prepare_router_data()
        ri1 = svc_helper.RouterInfo(router1['id'], router=router1)

        # Router #2 is like #1, except with some different IDs
        router2 = copy.deepcopy(router1)
        router2['id'] = _uuid()
        ri2 = svc_helper.RouterInfo(router2['id'], router=router2)
        h_info1 = create_hosting_info()
        h_info2 = copy.deepcopy(h_info1)
        h_info2['vrf_id'] = _uuid()
        ri1.router['hosting_info'] = h_info1
        ri2.router['hosting_info'] = h_info2

        ex_gw_port1 = ri1.router.get('gw_port')
        ex_gw_port2 = ri2.router.get('gw_port')
        ex_gw_port1['hosting_info'] = h_info1
        ex_gw_port2['hosting_info'] = h_info2
        vrf1 = ri1.router['hosting_info']['vrf_id']
        vrf2 = ri2.router['hosting_info']['vrf_id']
        driver._get_vrf_name = mock.Mock(return_value=vrf1)
        self.routing_helper._process_router(ri1)

        driver.internal_network_added.assert_called_with(
            ri1, ports[0])
        driver.enable_internal_network_NAT.assert_called_with(
            ri1, ports[0], ex_gw_port1)

        v_n_r_dict = self.routing_helper._router_ids_by_vrf_and_ext_net
        network_name = h_info1['network_name']
        self.assertEqual(1, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf1]))
        self.assertEqual(1, len(v_n_r_dict[vrf1][network_name]))

        driver.internal_network_added.reset_mock()
        driver.enable_internal_network_NAT.reset_mock()

        driver._get_vrf_name = mock.Mock(return_value=vrf2)
        self.routing_helper._process_router(ri2)
        # remove change_details as they are recursive and prevent comparison
        del ports[0]['change_details']
        del ri2.internal_ports[0]['change_details']
        driver.internal_network_added.assert_called_with(
            ri2, ports[0])
        driver.enable_internal_network_NAT.assert_called_with(
            ri2, ports[0], ex_gw_port2)
        network_name = h_info1['network_name']
        self.assertEqual(2, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf1]))
        self.assertEqual(1, len(v_n_r_dict[vrf2]))
        self.assertEqual(1, len(v_n_r_dict[vrf1][network_name]))
        self.assertEqual(1, len(v_n_r_dict[vrf2][network_name]))

        del ri1.router[bc.constants.INTERFACE_KEY]
        driver._get_vrf_name = mock.Mock(return_value=vrf1)
        self.routing_helper._process_router(ri1)

        driver.internal_network_removed.assert_called_with(
            ri1, ports[0], itfc_deleted=True)
        driver.disable_internal_network_NAT.assert_called_with(
            ri1, ports[0], ex_gw_port1, itfc_deleted=True)
        self.assertEqual(1, len(v_n_r_dict))
        self.assertFalse(v_n_r_dict.get(vrf1))
        self.assertEqual(1, len(v_n_r_dict[vrf2]))
        self.assertEqual(1, len(v_n_r_dict[vrf2][network_name]))

        driver.internal_network_removed.reset_mock()
        driver.disable_internal_network_NAT.reset_mock()

        del ri2.router[bc.constants.INTERFACE_KEY]
        driver._get_vrf_name = mock.Mock(return_value=vrf2)
        self.routing_helper._process_router(ri2)
        # remove change_details as they are recursive and prevent comparison
        del ports[0]['change_details']
        del driver.internal_network_removed.call_args[0][1]['change_details']
        driver.internal_network_removed.assert_called_with(
            ri2, ports[0], itfc_deleted=True)
        driver.disable_internal_network_NAT.assert_called_with(
            ri2, ports[0], ex_gw_port2, itfc_deleted=True)
        self.assertEqual({}, v_n_r_dict)

    def test_process_router_2_rids_1_vrf_2_networks(self):
        driver = _mock_driver_and_hosting_device(self.routing_helper)
        self._set_driver_port_mocks(driver)

        router1, ports = self.prepare_router_data()
        ri1 = svc_helper.RouterInfo(router1['id'], router=router1)

        # Router #2 is like #1, except with different IDs and host info
        router2 = copy.deepcopy(router1)
        router2['id'] = _uuid()
        ri2 = svc_helper.RouterInfo(router2['id'], router=router2)
        h_info1 = create_hosting_info()
        h_info2 = create_hosting_info(vrf=h_info1['vrf_id'],
            net_name=TEST_NET2, vlan=TEST_VLAN2, gw_ip=TEST_GW_IP2,
            cidr=TEST_CIDR2)
        ri1.router['hosting_info'] = h_info1
        ri2.router['hosting_info'] = h_info2

        ex_gw_port1 = ri1.router.get('gw_port')
        ex_gw_port2 = ri2.router.get('gw_port')
        ex_gw_port1['hosting_info'] = h_info1
        ex_gw_port2['hosting_info'] = h_info2
        network_name1 = h_info1['network_name']
        network_name2 = h_info2['network_name']
        vrf = ri1.router['hosting_info']['vrf_id']
        driver._get_vrf_name = mock.Mock(return_value=vrf)
        self.routing_helper._process_router(ri1)

        driver.internal_network_added.assert_called_with(
            ri1, ports[0])
        driver.enable_internal_network_NAT.assert_called_with(
            ri1, ports[0], ex_gw_port1)

        v_n_r_dict = self.routing_helper._router_ids_by_vrf_and_ext_net
        self.assertEqual(1, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf]))
        self.assertEqual(1, len(v_n_r_dict[vrf][network_name1]))

        driver.internal_network_added.reset_mock()
        driver.enable_internal_network_NAT.reset_mock()

        self.routing_helper._process_router(ri2)
        # remove change_details as they are recursive and prevent comparison
        del ports[0]['change_details']
        del ri2.internal_ports[0]['change_details']
        driver.internal_network_added.assert_called_with(
            ri2, ports[0])
        driver.enable_internal_network_NAT.assert_called_with(
            ri2, ports[0], ex_gw_port2)
        self.assertEqual(1, len(v_n_r_dict))
        self.assertEqual(2, len(v_n_r_dict[vrf]))
        self.assertEqual(1, len(v_n_r_dict[vrf][network_name1]))
        self.assertEqual(1, len(v_n_r_dict[vrf][network_name2]))

        del ri1.router[bc.constants.INTERFACE_KEY]
        self.routing_helper._process_router(ri1)

        driver.internal_network_removed.assert_called_with(
            ri1, ports[0], itfc_deleted=True)
        driver.disable_internal_network_NAT.assert_called_with(
            ri1, ports[0], ex_gw_port1, itfc_deleted=True)
        self.assertEqual(1, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf]))
        self.assertFalse(v_n_r_dict[vrf].get(network_name1))
        self.assertEqual(1, len(v_n_r_dict[vrf][network_name2]))

        driver.internal_network_removed.reset_mock()
        driver.disable_internal_network_NAT.reset_mock()

        del ri2.router[bc.constants.INTERFACE_KEY]
        self.routing_helper._process_router(ri2)

        # remove change_details as they are recursive and prevent comparison
        del ports[0]['change_details']
        del driver.internal_network_removed.call_args[0][1]['change_details']
        driver.internal_network_removed.assert_called_with(
            ri2, ports[0], itfc_deleted=True)
        driver.disable_internal_network_NAT.assert_called_with(
            ri2, ports[0], ex_gw_port2, itfc_deleted=True)
        self.assertEqual({}, v_n_r_dict)

    def test_process_router_2_rids_2_vrfs_2_networks(self):
        driver = _mock_driver_and_hosting_device(self.routing_helper)
        self._set_driver_port_mocks(driver)

        router1, ports = self.prepare_router_data()
        ri1 = svc_helper.RouterInfo(router1['id'], router=router1)

        # Router #2 is like #1, except with different IDs and host info
        router2 = copy.deepcopy(router1)
        router2['id'] = _uuid()
        ri2 = svc_helper.RouterInfo(router2['id'], router=router2)
        h_info1 = create_hosting_info()
        h_info2 = create_hosting_info(net_name=TEST_NET2,
            vlan=TEST_VLAN2, gw_ip=TEST_GW_IP2, cidr=TEST_CIDR2)
        ri1.router['hosting_info'] = h_info1
        ri2.router['hosting_info'] = h_info2

        ex_gw_port1 = ri1.router.get('gw_port')
        ex_gw_port2 = ri2.router.get('gw_port')
        ex_gw_port1['hosting_info'] = h_info1
        ex_gw_port2['hosting_info'] = h_info2
        vrf1 = ri1.router['hosting_info']['vrf_id']
        vrf2 = ri2.router['hosting_info']['vrf_id']
        network_name1 = h_info1['network_name']
        network_name2 = h_info2['network_name']
        driver._get_vrf_name = mock.Mock(return_value=vrf1)
        self.routing_helper._process_router(ri1)

        driver.internal_network_added.assert_called_with(
            ri1, ports[0])
        driver.enable_internal_network_NAT.assert_called_with(
            ri1, ports[0], ex_gw_port1)

        v_n_r_dict = self.routing_helper._router_ids_by_vrf_and_ext_net
        self.assertEqual(1, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf1]))
        self.assertEqual(1, len(v_n_r_dict[vrf1][network_name1]))

        driver.internal_network_added.reset_mock()
        driver.enable_internal_network_NAT.reset_mock()
        driver._get_vrf_name = mock.Mock(return_value=vrf2)

        self.routing_helper._process_router(ri2)
        # remove change_details as they are recursive and prevent comparison
        del ports[0]['change_details']
        del ri2.internal_ports[0]['change_details']
        driver.internal_network_added.assert_called_with(
            ri2, ports[0])
        driver.enable_internal_network_NAT.assert_called_with(
            ri2, ports[0], ex_gw_port2)
        self.assertEqual(2, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf1]))
        self.assertEqual(1, len(v_n_r_dict[vrf2]))
        self.assertEqual(1, len(v_n_r_dict[vrf1][network_name1]))
        self.assertEqual(1, len(v_n_r_dict[vrf2][network_name2]))

        del ri1.router[bc.constants.INTERFACE_KEY]
        driver._get_vrf_name = mock.Mock(return_value=vrf1)
        self.routing_helper._process_router(ri1)

        driver.internal_network_removed.assert_called_with(
            ri1, ports[0], itfc_deleted=True)
        driver.disable_internal_network_NAT.assert_called_with(
            ri1, ports[0], ex_gw_port1, itfc_deleted=True)
        self.assertEqual(1, len(v_n_r_dict))
        self.assertEqual(1, len(v_n_r_dict[vrf2]))
        self.assertFalse(v_n_r_dict.get(vrf1))
        self.assertEqual(1, len(v_n_r_dict[vrf2][network_name2]))

        driver.internal_network_removed.reset_mock()
        driver.disable_internal_network_NAT.reset_mock()

        del ri2.router[bc.constants.INTERFACE_KEY]
        driver._get_vrf_name = mock.Mock(return_value=vrf2)
        self.routing_helper._process_router(ri2)

        # remove change_details as they are recursive and prevent comparison
        del ports[0]['change_details']
        del driver.internal_network_removed.call_args[0][1]['change_details']
        driver.internal_network_removed.assert_called_with(
            ri2, ports[0], itfc_deleted=True)
        driver.disable_internal_network_NAT.assert_called_with(
            ri2, ports[0], ex_gw_port2, itfc_deleted=True)
        self.assertEqual({}, v_n_r_dict)



# Copyright (c) 2017 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import cb_events as events
from networking_cisco.backwards_compatibility import cb_registry as registry

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import (
    nexus_helpers as nexus_help)

LOG = log.getLogger(__name__)


class NexusTrunkHandler(object):
    """Cisco Nexus Trunk Handler.

    This class contains methods called by the trunk infrastruture
    to be processed by the cisco_nexus MD.
    """

    def __init__(self):
        self.plugin = bc.get_plugin()

    def _unbind_subport(self, context, port_id, status):
        self.plugin.update_port(context, port_id,
                                {'port':
                                 {bc.portbindings.HOST_ID: None,
                                  'status': status}})

    def trunk_update_postcommit(self, resource, event, trunk_plugin, payload):
        current_trunk_data = payload.current_trunk.to_dict()
        trunkport = self.plugin.get_port(
            payload.context, current_trunk_data['port_id'])

        if (nexus_help.is_baremetal(trunkport) and
            current_trunk_data['status'] != bc.constants.PORT_STATUS_ACTIVE):
            for subport in current_trunk_data['sub_ports']:
                self._unbind_subport(payload.context, subport['port_id'],
                                     current_trunk_data['status'])

    def subport_postcommit(self, resource, event, trunk_plugin, payload):
        trunkport = self.plugin.get_port(
            payload.context, payload.current_trunk.port_id)

        if (nexus_help.is_baremetal(trunkport) and
            trunkport['status'] == bc.constants.PORT_STATUS_ACTIVE):
            host_id = trunkport.get(bc.dns.DNSNAME)
            subport = payload.subports[0]
            trunk_subport = subport.to_dict()

            # Set the subport port attributes to match the parent port.
            if event == events.AFTER_CREATE:
                self.plugin.update_port(
                    payload.context, trunk_subport['port_id'],
                    {'port':
                     {bc.portbindings.HOST_ID: host_id,
                      'device_owner': bc.trunk_consts.TRUNK_SUBPORT_OWNER}})
            elif event == events.AFTER_DELETE:
                self._unbind_subport(
                    payload.context, trunk_subport['port_id'],
                    bc.constants.PORT_STATUS_DOWN)

            # Trunk drivers are responsible for setting the trunk
            # status. Use the trunk parent port's status.
            trunk_obj = bc.trunk_objects.Trunk.get_object(
                payload.context, id=payload.trunk_id)
            trunk_obj.update(status=trunkport['status'])


class NexusTrunkDriver(bc.trunk_base.DriverBase):
    """Cisco Nexus Trunk Driver.

    This class contains methods required to work with the trunk infrastruture.
    """
    @property
    def is_loaded(self):
        try:
            return (const.CISCO_NEXUS_ML2_MECH_DRIVER_V2 in
                    cfg.CONF.ml2.mechanism_drivers)
        except cfg.NoSuchOptError:
            return False

    def register(self, resource, event, trigger, **kwargs):
        super(NexusTrunkDriver, self).register(
            resource, event, trigger, **kwargs)
        self._handler = NexusTrunkHandler()

        registry.subscribe(self._handler.trunk_update_postcommit,
                           bc.trunk_consts.TRUNK, events.AFTER_UPDATE)
        for event in (events.AFTER_CREATE, events.AFTER_DELETE):
            registry.subscribe(self._handler.subport_postcommit,
                               bc.trunk_consts.SUBPORTS, event)

    @classmethod
    def create(cls):
        SUPPORTED_INTERFACES = (
            bc.portbindings.VIF_TYPE_OTHER,
        )

        SUPPORTED_SEGMENTATION_TYPES = (
            bc.trunk_consts.VLAN,
        )

        return cls(const.CISCO_NEXUS_ML2_MECH_DRIVER_V2,
                   SUPPORTED_INTERFACES,
                   SUPPORTED_SEGMENTATION_TYPES,
                   None,
                   can_trunk_bound_port=True)



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_cisco.plugins.cisco.cfg_agent.device_drivers.asr1k import (
    asr1k_cfg_syncer)
from networking_cisco.plugins.cisco.common.htparser import HTParser
from neutron.tests import base


CFG = """
!
interface TenGigabitEthernet0/0/0
 ip address 1.1.2.1 255.255.255.0
 no cdp enable
!
interface Serial1/0
 no ip address
!
ip access-list standard neutron_acl_2033_5c409ae1
 permit 10.10.0.0 0.0.255.255
    dummy-input
 dummy-input2
!
interface Port-channel11.2000
 description OPENSTACK_NEUTRON_EXTERNAL_INTF
 encapsulation dot1Q 2000
 ip address 15.15.6.117 255.255.0.0
 ip nat outside
 standby delay minimum 30 reload 60
 standby version 2
 standby 1064 ip 15.15.6.116
 standby 1064 timers 1 3
 standby 1064 priority 97
 standby 1064 name neutron-hsrp-1064-2000
 !
interface Port-channel11.2033
 description OPENSTACK_NEUTRON_INTF
 encapsulation dot1Q 2033
 vrf forwarding nrouter-6c5541
 ip address 10.10.0.12 255.255.0.0
 ip nat inside

interface Port-channel11.2055
 ip address 20.20.20.1 255.255.255.0
"""

FIND_LINES = "find_lines"
FIND_CHILDREN = "find_children"
FIND_OBJECTS = "find_objects"
RE_SEARCH_CHILDREN = "re_search_children"
ACL_CHILD_REGEX = asr1k_cfg_syncer.ACL_CHILD_REGEX


class TestHTParser(base.BaseTestCase):
    def setUp(self):
        super(TestHTParser, self).setUp()
        self.cfg = [x for x in CFG.splitlines()]

    def _execute(self, function_name, linespec):
        return getattr(HTParser(self.cfg), function_name)(linespec)

    def test_find_lines_singleline(self):
        linespec = "^interface TenGigabitEthernet0/0/0"
        expected = ["interface TenGigabitEthernet0/0/0"]
        actual = self._execute(FIND_LINES, linespec)
        self.assertEqual(expected, actual)

    def test_find_lines_mutliline(self):
        extra_cfg = [
            'interface TenGigabitEthernet0/0/0',
            '    no ip address',
            '    channel-group 10 mode active',
            '    !'
        ]
        self.cfg.extend(extra_cfg)
        linespec = "^interface TenGigabitEthernet0/0/0"
        exp = ["interface TenGigabitEthernet0/0/0",
               "interface TenGigabitEthernet0/0/0"]
        act = self._execute(FIND_LINES, linespec)
        self.assertEqual(exp, act)

    def test_find_lines_similiarlines(self):
        extra_cfg = [
            'interface TenGigabitEthernet0/0/1',
            '    no ip address',
            '    channel-group 10 mode active',
            '    !'
        ]
        self.cfg.extend(extra_cfg)
        linespec = "^interface TenGigabitEthernet0/0/0"
        exp = ["interface TenGigabitEthernet0/0/0"]
        act = self._execute(FIND_LINES, linespec)
        self.assertEqual(exp, act)

    def test_find_lines_multiple(self):
        extra_cfg = [
            'interface TenGigabitEthernet0/0/1',
            '    no ip address',
            '    channel-group 10 mode active',
            '    !'
        ]
        self.cfg.extend(extra_cfg)
        linespec = "^interface TenGigabitEthernet"
        exp = ["interface TenGigabitEthernet0/0/0",
               "interface TenGigabitEthernet0/0/1"]
        act = self._execute(FIND_LINES, linespec)
        self.assertEqual(exp, act)

    def test_find_lines_vrf_def(self):
        extra_cfg = [
            'vrf definition nrouter-bf7c12',
            ' !',
            ' address-family ipv4',
            ' exit-address-family',
            ' !',
            ' address-family ipv6',
            ' exit-address-family',
            ' !',
        ]
        self.cfg.extend(extra_cfg)
        linespec = "^vrf definition"
        exp = ["vrf definition nrouter-bf7c12"]
        act = self._execute(FIND_LINES, linespec)
        self.assertEqual(exp, act)

    def test_find_children_interface(self):
        linespec = "^interface TenGigabitEthernet0/0/0"
        exp = ["interface TenGigabitEthernet0/0/0",
               " ip address 1.1.2.1 255.255.255.0",
               " no cdp enable"]
        act = self._execute(FIND_CHILDREN, linespec)
        self.assertEqual(exp, act)

    def test_find_children_acl(self):
        linespec = "ip access-list standard neutron_acl_2033_5c409ae1"
        exp = ['ip access-list standard neutron_acl_2033_5c409ae1',
               ' permit 10.10.0.0 0.0.255.255',
               ' dummy-input2']
        act = self._execute(FIND_CHILDREN, linespec)
        self.assertEqual(exp, act)

    def test_find_children_acl_multiline(self):
        extra_cfg = [
            'ip access-list standard neutron_acl_1234_5678',
            '    permit 10.10.0.0 0.0.255.255'
        ]
        self.cfg.extend(extra_cfg)
        linespec = "ip access-list standard neutron_acl_*"
        exp = ['ip access-list standard neutron_acl_2033_5c409ae1',
               ' permit 10.10.0.0 0.0.255.255',
               ' dummy-input2',
               'ip access-list standard neutron_acl_1234_5678',
               '    permit 10.10.0.0 0.0.255.255']
        act = self._execute(FIND_CHILDREN, linespec)
        self.assertEqual(exp, act)

    def test_find_children_acl_no_find(self):
        linespec = "ip access-list standard neutron_acl_xyz"
        exp = []
        act = self._execute(FIND_CHILDREN, linespec)
        self.assertEqual(exp, act)

    def test_find_children_interface_spec(self):
        linespec = "^interface Port-channel11.2055"
        exp = ['interface Port-channel11.2055',
               ' ip address 20.20.20.1 255.255.255.0']
        act = self._execute(FIND_CHILDREN, linespec)
        self.assertEqual(exp, act)

    # Object based tests
    def test_find_objects(self):
        linespec = "^interface TenGigabitEthernet0/0/0"
        exp = ['interface TenGigabitEthernet0/0/0',
               ' ip address 1.1.2.1 255.255.255.0',
               ' no cdp enable']
        act = self._execute(FIND_OBJECTS, linespec)
        self.assertEqual(1, len(act))
        self.assertEqual(exp, act[0].str_list())

    def test_find_objects_no_find(self):
        linespec = "^interface TenGigabitEthernet0/0/1"
        exp = []
        act = self._execute(FIND_OBJECTS, linespec)
        self.assertEqual(exp, act)

    def test_re_search_children_interface(self):
        linespec = "^interf"
        child_linespec = "\s*no ip*"
        exp = ['interface Serial1/0', ' no ip address']
        parse = HTParser(self.cfg)
        act = [obj for obj in parse.find_objects(linespec)
               if obj.re_search_children(child_linespec)]
        self.assertEqual(1, len(act))
        self.assertEqual(exp, act[0].str_list())

    def test_re_search_children_acl(self):
        linespec = "^ip access"
        child_linespec = ACL_CHILD_REGEX
        exp = ['ip access-list standard neutron_acl_2033_5c409ae1',
               ' permit 10.10.0.0 0.0.255.255',
               ' dummy-input2']
        parse = HTParser(self.cfg)
        act = [obj for obj in parse.find_objects(linespec)
               if obj.re_search_children(child_linespec)]
        self.assertEqual(1, len(act))
        self.assertEqual(exp, act[0].str_list())

    def test_re_match_nat(self):
        linespec = "^interf"
        child_linespec = "\s*ip nat inside"
        match_regex = "^interface (\S+)"
        exp = 'Port-channel11.2033'
        parse = HTParser(self.cfg)
        act = [obj for obj in parse.find_objects(linespec)
               if obj.re_search_children(child_linespec)]
        self.assertEqual(1, len(act))
        self.assertEqual(exp, act[0].re_match(match_regex))

    def test_build_indent_based_list__singleline_noindent(self):
        LINE = 'cfg_line'
        cfg = LINE
        parser = HTParser(cfg)
        parser.find_objects("fake_object")
        self.assertEqual([(0, LINE)], parser._indent_list)

    def test_build_indent_based_list__multiline_noindent(self):
        LINE = 'cfg_line'
        cfg = LINE + '\n' + LINE
        parser = HTParser(cfg)
        parser.find_objects("fake_object")
        res = [(0, LINE), (0, LINE)]
        self.assertEqual(res, parser._indent_list)

    def test_build_indent_based_list__singleline_indent(self):
        LINE = '  cfg_line'
        cfg = LINE
        parser = HTParser(cfg)
        parser.find_objects("fake_object")
        self.assertEqual([(2, LINE)], parser._indent_list)

    def test_build_indent_based_list__singleline_comment(self):
        LINE = '!'
        cfg = LINE
        parser = HTParser(cfg)
        parser.find_objects("fake_object")
        self.assertEqual([], parser._indent_list)

    def test_build_indent_based_list__multiline_indent(self):
        LINE = '  cfg_line'
        cfg = LINE + '\n' + LINE
        parser = HTParser(cfg)
        parser.find_objects("fake_object")
        res = [(2, LINE), (2, LINE)]
        self.assertEqual(res, parser._indent_list)

    def test_init__multiline_blank(self):
        LINE = ""
        cfg = LINE + "\n" + LINE + "\n" + LINE
        parser = HTParser(cfg)
        parser.find_objects("fake_object")
        self.assertEqual([], parser._indent_list)

    def test_init__multiline_mixed(self):
        LINE = " cfg_line"
        cfg = LINE + "\n" + "  \n" + LINE
        parser = HTParser(cfg)
        parser.find_objects("fake_object")
        res = [(1, ' cfg_line'), (1, ' cfg_line')]
        self.assertEqual(res, parser._indent_list)

    def test_init__multiline_mixed_different_indent(self):
        LINE = "cfg_line"
        cfg = "  " + LINE + "\n" + "  \n" + "   " + LINE
        parser = HTParser(cfg)
        parser.find_objects("fake_object")
        res = [(2, '  cfg_line'), (3, '   cfg_line')]
        self.assertEqual(res, parser._indent_list)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutronclient._i18n import _
from neutronclient.common import extension


class PolicyProfile(extension.NeutronClientExtension):
    resource = 'policy_profile'
    resource_plural = '%ss' % resource
    object_path = '/%s' % resource_plural
    resource_path = '/%s/%%s' % resource_plural
    versions = ['2.0']
    allow_names = True


class PolicyProfileList(extension.ClientExtensionList, PolicyProfile):
    """List policy profiles that belong to a given tenant."""

    shell_command = 'cisco-policy-profile-list'
    list_columns = ['id', 'name']
    pagination_support = True
    sorting_support = True


class PolicyProfileShow(extension.ClientExtensionShow, PolicyProfile):
    """Show information of a given policy profile."""

    shell_command = 'cisco-policy-profile-show'


class UpdatePolicyProfile(extension.ClientExtensionUpdate, PolicyProfile):
    """Update policy profile's information."""

    shell_command = 'cisco-policy-profile-update'

    def get_parser(self, prog_name):
        parser = super(UpdatePolicyProfile, self).get_parser(prog_name)
        parser.add_argument("--add-tenant",
                            action='append',
                            dest='add_tenant',
                            help=_("Add tenant to the policy profile."))
        parser.add_argument("--remove-tenant",
                            action='append',
                            dest='remove_tenant',
                            help=_("Remove tenant from the policy profile."))
        return parser

    def args2body(self, parsed_args):
        body = {self.resource: {}}
        if parsed_args.add_tenant:
            body[self.resource]['add_tenant'] = parsed_args.add_tenant
        if parsed_args.remove_tenant:
            body[self.resource]['remove_tenant'] = parsed_args.remove_tenant
        return body



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Service VDP 2.2 constants."""

MINIMUM_VDP22_VERSION = "0.9.45"
VDP_MGRID = 0
VDP_TYPEID = 0
VDP_TYPEID_VER = 0
VDP_VSIFRMT_UUID = 5
VDP_FILTER_VID = 1
VDP_FILTER_MACVID = 2
VDP_FILTER_GIDVID = 3
VDP_FILTER_GIDMACVID = 4
VDP_SYNC_TIMEOUT = 15
CALLBACK_THRESHOLD = 5

verify_failure_reason = "vsi_id mismatch, queried %s, returned %s"
retrieve_failure_reason = "Unable to retrieve failure, reply %s"
mode_failure_reason = "Incorrect Reply,no mode information found: %s"
filter_failure_reason = "Incorrect Reply,no filter information found: %s"
multiple_filter_failure_reason = \
    "Incorrect Reply,multiple filter information found: %s"
format_failure_reason = "Reply not formatted correctly: %s"
hints_failure_reason = "Incorrect Reply,no hints information found: %s"
multiple_hints_failure_reason = \
    "Incorrect Reply,multiple hints information found: %s"
nonzero_hints_failure = "Non-zero hints, value %d"
vsi_mismatch_failure_reason = "VSIID Reply mis-match req vsi %s reply vsi %s"
mac_mismatch_failure_reason = \
    "VSIID MAC Reply mis-match req mac %s reply mac %s"



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sys

import mock

from neutronclient.common import exceptions
from neutronclient import shell
from neutronclient.tests.unit import test_cli20

from networking_cisco.neutronclient import routertype


class CLITestV20RouterType(test_cli20.CLITestV20Base):

    def setUp(self):
        # need to mock before super because extensions loaded on instantiation
        self._mock_extension_loading()
        super(CLITestV20RouterType, self).setUp()
        self.non_admin_status_resources.append('routertype')

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        contrib = mock.patch(ext_pkg + '._discover_via_entry_points').start()
        contrib.return_value = [("routertype", routertype)]
        return contrib

    def test_ext_cmd_loaded(self):
        shell.NeutronShell('2.0')
        ext_cmd = {'cisco-router-type-list': routertype.RouterTypeList,
                   'cisco-router-type-create': routertype.RouterTypeCreate,
                   'cisco-router-type-update': routertype.RouterTypeUpdate,
                   'cisco-router-type-delete': routertype.RouterTypeDelete,
                   'cisco-router-type-show': routertype.RouterTypeShow}
        self.assertDictContainsSubset(ext_cmd, shell.COMMANDS['2.0'])

    def test_ext_cmd_help_doc_with_extension_name(self):
        shell.NeutronShell('2.0')
        ext_cmd = {'cisco-router-type-list': routertype.RouterTypeList,
                   'cisco-router-type-create': routertype.RouterTypeCreate,
                   'cisco-router-type-update': routertype.RouterTypeUpdate,
                   'cisco-router-type-delete': routertype.RouterTypeDelete,
                   'cisco-router-type-show': routertype.RouterTypeShow}
        self.assertDictContainsSubset(ext_cmd, shell.COMMANDS['2.0'])
        for item in ext_cmd:
            cmdcls = shell.COMMANDS['2.0'].get(item)
            self.assertTrue(cmdcls.__doc__.startswith("[routertype]"))

    def test_create_router_type(self):
        """Create router type."""
        resource = 'routertype'
        cmd = routertype.RouterTypeCreate(test_cli20.MyApp(sys.stdout), None)
        template = 'Template 1'
        scheduler = 'my.scheduler:class_name'
        plugin_driver = 'my.plugin.driver:class_name'
        svc_helper = 'my.service.helper:class_name'
        agent_driver = 'my.agent.driver:class_name'
        myid = 'myid'
        args = [template, scheduler, plugin_driver, svc_helper, agent_driver]
        position_names = ['template_id', 'scheduler', 'driver',
                          'cfg_agent_service_helper', 'cfg_agent_driver']
        position_values = [template, scheduler, plugin_driver, svc_helper,
                           agent_driver]
        self._test_create_resource(resource, cmd, None, myid, args,
                                   position_names, position_values)

    def test_create_router_type_id(self):
        """Create router type: --id this_id myname."""
        resource = 'routertype'
        cmd = routertype.RouterTypeCreate(test_cli20.MyApp(sys.stdout), None)
        template = 'Template 1'
        scheduler = 'my.scheduler:class_name'
        plugin_driver = 'my.plugin.driver:class_name'
        svc_helper = 'my.service.helper:class_name'
        agent_driver = 'my.agent.driver:class_name'
        myid = 'myid'
        args = ['--id', myid, template, scheduler, plugin_driver,
                svc_helper, agent_driver]
        position_names = ['template_id', 'scheduler', 'driver',
                          'cfg_agent_service_helper', 'cfg_agent_driver']
        position_values = [template, scheduler, plugin_driver, svc_helper,
                           agent_driver]
        self._test_create_resource(resource, cmd, None, myid, args,
                                   position_names, position_values,
                                   id=myid)

    def test_create_router_type_tenant(self):
        """Create router type: --tenant_id tenantid myname."""
        resource = 'routertype'
        cmd = routertype.RouterTypeCreate(test_cli20.MyApp(sys.stdout), None)
        template = 'Template 1'
        scheduler = 'my.scheduler:class_name'
        plugin_driver = 'my.plugin.driver:class_name'
        svc_helper = 'my.service.helper:class_name'
        agent_driver = 'my.agent.driver:class_name'
        myid = 'myid'
        args = ['--tenant_id', 'tenantid', template, scheduler, plugin_driver,
                svc_helper, agent_driver]
        position_names = ['template_id', 'scheduler', 'driver',
                          'cfg_agent_service_helper', 'cfg_agent_driver']
        position_values = [template, scheduler, plugin_driver, svc_helper,
                           agent_driver]
        self._test_create_resource(resource, cmd, None, myid, args,
                                   position_names, position_values,
                                   tenant_id='tenantid')

    def _test_create_router_type_optional_args(
            self, name=None, desc=None, ha_enabled=None, haenabled=None,
            unshared=None, slot_need=None, slotneed=None):
        resource = 'routertype'
        cmd = routertype.RouterTypeCreate(test_cli20.MyApp(sys.stdout),
                                          None)
        myid = 'myid'
        template = 'Template 1'
        scheduler = 'my.scheduler:class_name'
        plugin_driver = 'my.plugin.driver:class_name'
        svc_helper = 'my.service.helper:class_name'
        agent_driver = 'my.agent.driver:class_name'
        args = []
        expected = {}
        if desc is not None:
            args += ['--description', desc]
            expected['description'] = desc
        if ha_enabled is not None:
            args += ['--ha_enabled']
            expected['ha_enabled_by_default'] = True
        if haenabled is not None:
            args += ['--ha-enabled']
            expected['ha_enabled_by_default'] = True
        if unshared is not None:
            args += ['--unshared']
            expected['shared'] = False
        if slot_need is not None:
            args += ['--slot_need', slot_need]
            expected['slot_need'] = slot_need
        if slotneed is not None:
            args += ['--slot-need', slotneed]
            expected['slot_need'] = slotneed
        position_names = ['template_id', 'scheduler', 'driver',
                          'cfg_agent_service_helper', 'cfg_agent_driver']
        position_values = [template, scheduler, plugin_driver, svc_helper,
                           agent_driver]
        for p_v in position_values:
            args.append(p_v)
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values,
                                   **expected)

    def test_create_router_type_name(self):
        self._test_create_router_type_optional_args('some name')

    def test_create_router_type_description(self):
        self._test_create_router_type_optional_args(desc='some description')

    def test_create_router_type_ha(self):
        self._test_create_router_type_optional_args(ha_enabled=True)
        self._test_create_router_type_optional_args(haenabled=True)

    def test_create_router_type_unshared(self):
        self._test_create_router_type_optional_args(unshared=False)

    def test_create_router_type_slots(self):
        self._test_create_router_type_optional_args(slot_need='5')
        self._test_create_router_type_optional_args(slotneed='5')

    def test_create_router_type_full(self):
        self._test_create_router_type_optional_args(
            'some name', desc='some description', ha_enabled=True,
            unshared=False, slot_need='5')
        self._test_create_router_type_optional_args(
            'some name', desc='some description', haenabled=True,
            unshared=False, slotneed='5')

    def test_list_router_types_detail(self):
        """list routers: -D."""
        resources = "routertypes"
        cmd = routertype.RouterTypeList(test_cli20.MyApp(sys.stdout), None)
        response_contents = [{
            self.id_field: 'myid1', 'name': 'router_type_1_name',
            'description': 'fast router', 'template_id': 'templ_id_1'}, {
            self.id_field: 'myid2', 'name': 'router_type_2_name',
            'description': 'faster router', 'template_id': 'templ_id_2'}]
        self._test_list_resources(resources, cmd, True,
                                  response_contents=response_contents)

    def test_list_router_types_sort(self):
        """list routertypes: --sort-key name --sort-key id --sort-key asc
        --sort-key desc
        """
        resources = "routertypes"
        cmd = routertype.RouterTypeList(test_cli20.MyApp(sys.stdout), None)
        response_contents = [{
            self.id_field: 'myid1', 'name': 'router_type_1_name',
            'description': 'fast router', 'template_id': 'templ_id_1'}, {
            self.id_field: 'myid2', 'name': 'router_type_2_name',
            'description': 'faster router', 'template_id': 'templ_id_2'}]
        self._test_list_resources(resources, cmd,
                                  sort_key=["name", "id"],
                                  sort_dir=["asc", "desc"],
                                  response_contents=response_contents)

    def test_list_router_types_limit(self):
        """list routertypes: -P."""
        resources = "routertypes"
        cmd = routertype.RouterTypeList(test_cli20.MyApp(sys.stdout), None)
        response_contents = [{
            self.id_field: 'myid1', 'name': 'router_type_1_name',
            'description': 'fast router', 'template_id': 'templ_id_1'}, {
            self.id_field: 'myid2', 'name': 'router_type_2_name',
            'description': 'faster router', 'template_id': 'templ_id_2'}]
        self._test_list_resources(resources, cmd, page_size=1000,
                                  response_contents=response_contents)

    def test_update_router_type_exception(self):
        """Update routertype: myid."""
        resource = 'routertype'
        cmd = routertype.RouterTypeUpdate(test_cli20.MyApp(sys.stdout), None)
        self.assertRaises(exceptions.CommandError, self._test_update_resource,
                          resource, cmd, 'myid', ['myid'], {})

    def _test_update_router_type(
            self, name=None, desc=None, ha_enabled=None, haenabled=None,
            ha_disabled=None, hadisabled=None, unshared=None, shared=None,
            slot_need=None, slotneed=None):
        resource = 'routertype'
        cmd = routertype.RouterTypeUpdate(test_cli20.MyApp(sys.stdout),
                                          None)
        myid = 'myid'
        args = [myid]
        expected = {}
        if name is not None:
            args += ['--name', name]
            expected['name'] = name
        if desc is not None:
            args += ['--description', desc]
            expected['description'] = desc
        if ha_enabled is not None:
            args += ['--ha_enabled']
            expected['ha_enabled_by_default'] = True
        if haenabled is not None:
            args += ['--ha-enabled']
            expected['ha_enabled_by_default'] = True
        if ha_disabled is not None:
            args += ['--ha_disabled']
            expected['ha_enabled_by_default'] = False
        if hadisabled is not None:
            args += ['--ha-disabled']
            expected['ha_enabled_by_default'] = False
        if shared is not None:
            args += ['--shared']
            expected['shared'] = True
        if unshared is not None:
            args += ['--unshared']
            expected['shared'] = False
        if slot_need is not None:
            args += ['--slot_need', slot_need]
            expected['slot_need'] = slot_need
        if slotneed is not None:
            args += ['--slot-need', slotneed]
            expected['slot_need'] = slotneed
        self._test_update_resource(resource, cmd, myid, args, expected)

    def test_update_router_type_name(self):
        """Update routertype: myid --name myname."""
        self._test_update_router_type(name='myname')

    def test_update_router_type_description(self):
        self._test_update_router_type(desc='some description')

    def test_update_router_type_ha(self):
        self._test_update_router_type(ha_enabled=True)
        self._test_update_router_type(haenabled=True)
        self._test_update_router_type(ha_disabled=True)
        self._test_update_router_type(hadisabled=True)

    def test_update_router_type_sharing(self):
        self._test_update_router_type(shared=True)
        self._test_update_router_type(unshared=True)

    def test_update_router_type_slots(self):
        self._test_update_router_type(slot_need='5')
        self._test_update_router_type(slotneed='5')

    def test_update_router_type_full(self):
        self._test_update_router_type(name='myname', desc='some description',
                                      ha_enabled=True, shared=True,
                                      slot_need='5')
        self._test_update_router_type(name='myname', desc='some description',
                                      haenabled=True, shared=True,
                                      slotneed='5')
        self._test_update_router_type(name='myname', desc='some description',
                                      ha_disabled=True, unshared=True,
                                      slot_need='5')
        self._test_update_router_type(name='myname', desc='some description',
                                      hadisabled=True, unshared=True,
                                      slotneed='5')

    def test_delete_router_type(self):
        """Delete routertype: myid."""
        resource = 'routertype'
        cmd = routertype.RouterTypeDelete(test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        args = [myid]
        self._test_delete_resource(resource, cmd, myid, args)

    def test_show_router_type(self):
        """Show routertype: myid."""
        resource = 'routertype'
        cmd = routertype.RouterTypeShow(test_cli20.MyApp(sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id, args,
                                 ['id', 'name'])



# Copyright (c) 2013-2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Implements a Nexus-OS NETCONF over SSHv2 API Client
"""

import re
import six
import time
import warnings

from debtcollector import removals
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import (
    exceptions as cexc)
from networking_cisco.ml2_drivers.nexus import (
    nexus_base_network_driver as basedrvr)
from networking_cisco.ml2_drivers.nexus import (
    nexus_snippets as snipp)

LOG = logging.getLogger(__name__)
warnings.simplefilter('always')


@removals.removed_class('CiscoNexusSshDriver',
                        message="CiscoNexusSshDriver replaced by "
                                "default CiscoNexusRestapiDriver.",
                        version="Cisco 5.1.0",
                        removal_version="Cisco 7.0.0")
class CiscoNexusSshDriver(basedrvr.CiscoNexusBaseDriver):
    """Nexus Driver Main Class."""
    def __init__(self, nexus_switches):
        super(CiscoNexusSshDriver, self).__init__(nexus_switches)
        self.ncclient = None
        self.connections = {}
        self.init_ssh_caching()
        LOG.debug("ML2 Nexus SSH Drivers initialized.")

    # Driver lock introduced to prevent replay thread and
    # transaction thread from closing each others
    # session before complete.
    @lockutils.synchronized('cisco-nexus-drvrlock')
    def keep_ssh_caching(self):
        self._close_ssh_session = False

    # Driver lock introduced to prevent replay thread and
    # transaction thread from closing each others
    # session before complete.
    @lockutils.synchronized('cisco-nexus-drvrlock')
    def init_ssh_caching(self):
        self._close_ssh_session = True if (
            cfg.CONF.ml2_cisco.never_cache_ssh_connection or
            (cfg.CONF.rpc_workers + cfg.CONF.api_workers) >=
            const.MAX_NEXUS_SSH_SESSIONS) else False

    def _import_client(self):
        """Import the NETCONF client (ncclient) module.

        The ncclient module is not installed as part of the normal Neutron
        distributions. It is imported dynamically in this module so that
        the import can be mocked, allowing unit testing without requiring
        the installation of ncclient.

        """
        return importutils.import_module('ncclient.manager')

    def _get_close_ssh_session(self):
        return self._close_ssh_session

    def _close_session(self, mgr, nexus_host):
        """Close the connection to the nexus switch."""
        starttime = time.time()
        if mgr:
            self.connections.pop(nexus_host, None)
            mgr.close_session()
            self.capture_and_print_timeshot(
                starttime, "close",
                switch=nexus_host)
        else:
            self.capture_and_print_timeshot(
                starttime, "nomgr_close",
                switch=nexus_host)

    # Driver lock introduced to prevent replay thread and
    # transaction thread from closing each others
    # session before complete.
    @lockutils.synchronized('cisco-nexus-drvrlock')
    def close_session(self, nexus_host):
        mgr = self._nxos_connect(nexus_host)
        self._close_session(mgr, nexus_host)

    # Driver lock introduced to prevent replay thread and
    # transaction thread from closing each others
    # session before complete.
    @lockutils.synchronized('cisco-nexus-drvrlock')
    def _get_config(self, nexus_host, filter='',
                    check_to_close_session=True):
        """Get Nexus Host Configuration

        :param nexus_host: IP address of switch
        :param filter: filter string in XML format
        :returns: Configuration requested in string format
        """

        # For loop added to handle stale ncclient handle after switch reboot.
        # If the attempt fails,
        #     close the session, save first exception
        #     loop back around
        #     try again
        #     then quit
        for retry_count in (1, 2):
            mgr = self._nxos_connect(nexus_host)
            starttime = time.time()
            try:
                data_xml = mgr.get(filter=('subtree', filter)).data_xml
            except Exception as e:
                self.capture_and_print_timeshot(
                    starttime, "geterr", retry_count, switch=nexus_host)
                try:
                    self._close_session(mgr, nexus_host)
                except Exception:
                    pass

                # if transaction is snipp.EXEC_GET_INVENTORY_SNIPPET,
                # don't retry since this is used as a ping to
                # validate connection and retry is already built
                # into replay code.
                if snipp.EXEC_GET_INVENTORY_SNIPPET == filter:
                    raise cexc.NexusConfigFailed(nexus_host=nexus_host,
                                                 config=filter,
                                                 exc=e)

                # if first try, save first exception and retry
                if retry_count == 1:
                    first_exc = e
                else:
                    raise cexc.NexusConfigFailed(nexus_host=nexus_host,
                                                 config=filter,
                                                 exc=first_exc)
            else:
                self.capture_and_print_timeshot(
                    starttime, "get",
                    switch=nexus_host)
                if check_to_close_session and self._get_close_ssh_session():
                    self._close_session(mgr, nexus_host)
                return data_xml

    # Driver lock introduced to prevent replay thread and
    # transaction thread from closing each others
    # session before complete.
    @lockutils.synchronized('cisco-nexus-drvrlock')
    def _edit_config(self, nexus_host, target='running', config='',
                     allowed_exc_strs=None, check_to_close_session=True):
        """Modify switch config for a target config type.

        :param nexus_host: IP address of switch to configure
        :param target: Target config type
        :param config: Configuration string in XML format
        :param allowed_exc_strs: Exceptions which have any of these strings
                                 as a subset of their exception message
                                 (str(exception)) can be ignored
        :param check_to_close_session: Set to False when configured to close
                                       the ssh session is not to be checked.

        :raises: NexusConfigFailed: if _edit_config() encountered an exception
                                    not containing one of allowed_exc_strs

        """
        if not allowed_exc_strs:
            allowed_exc_strs = []

        # For loop added to handle stale ncclient handle after switch reboot.
        # If the attempt fails and not an allowed exception,
        #     close the session, save first exception
        #     loop back around
        #     try again
        #     then quit
        for retry_count in (1, 2):
            mgr = self._nxos_connect(nexus_host)
            LOG.debug("NexusDriver edit config for host %s: %s",
                      nexus_host, config)
            starttime = time.time()
            try:
                mgr.edit_config(target=target, config=config)
                break
            except Exception as e:
                self.capture_and_print_timeshot(
                    starttime, "editerr", retry_count, switch=nexus_host)

                for exc_str in allowed_exc_strs:
                    if exc_str in six.u(str(e)):
                        return
                try:
                    self._close_session(mgr, nexus_host)
                except Exception:
                    pass
                if retry_count == 1:
                    first_exc = e
                else:
                    # Raise a Neutron exception. Include a description of
                    # the original ncclient exception.
                    raise cexc.NexusConfigFailed(nexus_host=nexus_host,
                                                 config=config,
                                                 exc=first_exc)

        self.capture_and_print_timeshot(
            starttime, "edit",
            switch=nexus_host)

        # if configured, close the ncclient ssh session.
        if check_to_close_session and self._get_close_ssh_session():
            self._close_session(mgr, nexus_host)

    def _nxos_connect(self, nexus_host):
        """Make SSH connection to the Nexus Switch."""

        starttime = time.time()
        if hasattr(self.connections.get(nexus_host), 'connected'):
            self.capture_and_print_timeshot(
                starttime, "cacheconnect",
                switch=nexus_host)
            return self.connections[nexus_host]

        if not self.ncclient:
            self.ncclient = self._import_client()
        nexus_ssh_port = int(self.nexus_switches[nexus_host][const.SSHPORT])
        nexus_user = self.nexus_switches[nexus_host][const.USERNAME]
        nexus_password = self.nexus_switches[nexus_host][const.PASSWORD]
        hostkey_verify = cfg.CONF.ml2_cisco.host_key_checks
        try:
            # With new ncclient version, we can pass device_params...
            man = self.ncclient.connect(
                host=nexus_host,
                port=nexus_ssh_port,
                username=nexus_user,
                password=nexus_password,
                hostkey_verify=hostkey_verify,
                timeout=30,
                device_params={
                    "name": "nexus",
                    'ssh_subsystem_name': 'xmlagent'})
        except Exception as e:
            # Raise a Neutron exception. Include a description of
            # the original ncclient exception.
            self.capture_and_print_timeshot(
                starttime, "connecterr",
                switch=nexus_host)
            raise cexc.NexusConnectFailed(nexus_host=nexus_host, exc=e)

        self.capture_and_print_timeshot(
            starttime, "connect",
            switch=nexus_host)
        self.connections[nexus_host] = man
        return self.connections[nexus_host]

    def create_xml_snippet(self, customized_config):
        """Create XML snippet.

        Creates the Proper XML structure for the Nexus Switch Configuration.
        and also does 'copy run start' if configured to do so.  This
        latter command allows configuration to persist on the switch after
        reboot.
        """
        if cfg.CONF.ml2_cisco.persistent_switch_config:
            customized_config += (snipp.EXEC_SAVE_CONF_SNIPPET)

        conf_xml_snippet = snipp.EXEC_CONF_SNIPPET % (customized_config)
        return conf_xml_snippet

    def get_interface_switch(self, nexus_host,
                             intf_type, interface):
        """Get the interface data from host.

        :param nexus_host: IP address of Nexus switch
        :param intf_type:  String which specifies interface type.
                           example: ethernet
        :param interface:  String indicating which interface.
                           example: 1/19
        :returns response:
        """

        confstr = snipp.EXEC_GET_INTF_SNIPPET % (intf_type, interface)
        starttime = time.time()
        response = self._get_config(nexus_host, confstr)
        self.capture_and_print_timeshot(starttime, "getif",
            switch=nexus_host)
        LOG.debug("GET call returned interface %(if_type)s %(interface)s "
            "config", {'if_type': intf_type, 'interface': interface})
        return response

    def initialize_baremetal_switch_interfaces(self, interfaces):
        """Initialize Nexus interfaces and for initial baremetal event."""

        self.initialize_all_switch_interfaces(interfaces)

    def initialize_all_switch_interfaces(self, interfaces, switch_ip=None,
                                         replay=True):
        """Configure Nexus interface and get port channel number.

        :param interfaces:  List of interfaces for a given switch.
                            ch_grp can be altered as last arg
                            to each interface. If no ch_grp,
                            this arg will be zero.
        :param switch_ip: IP address of Nexus switch
        :param replay: Whether in replay path
        """

        if not interfaces:
            return

        starttime = time.time()
        ifs = []
        for i, (nexus_host, intf_type, nexus_port, is_native,
            ch_grp_saved) in enumerate(interfaces):
            response = self.get_interface_switch(
                           nexus_host, intf_type, nexus_port)
            if ch_grp_saved == 0:
                # Collect the port-channel number from response
                mo = re.search("channel-group\s(\d*)\s", response)
                try:
                    ch_grp = int(mo.group(1))
                except Exception:
                    ch_grp = 0
                #substitute content of ch_grp
                no_chgrp_len = len(interfaces[i]) - 1
                interfaces[i] = interfaces[i][:no_chgrp_len] + (ch_grp,)
            else:
                ch_grp = ch_grp_saved
            if ch_grp is not 0:
                # if channel-group exists, init port-channel
                # instead of the provided ethernet interface
                intf_type = 'port-channel'
                nexus_port = str(ch_grp)
            if (response and
               "switchport trunk allowed vlan" in response):
                pass
            else:
                ifs.append(self._build_intf_confstr(
                           snipp.CMD_INT_VLAN_SNIPPET,
                           intf_type, nexus_port, 'None'))
        if ifs:
            confstr = self.create_xml_snippet(''.join(ifs))
            self._edit_config(nexus_host, target='running',
                              config=confstr)
        self.capture_and_print_timeshot(
            starttime, "get_allif",
            switch=nexus_host)

    def get_nexus_type(self, nexus_host):
        """Given the nexus host, get the type of Nexus switch.

        :param nexus_host: IP address of Nexus switch
        :returns: Nexus type
        """

        confstr = snipp.EXEC_GET_INVENTORY_SNIPPET
        starttime = time.time()
        response = self._get_config(nexus_host, confstr)
        self.capture_and_print_timeshot(
            starttime, "gettype",
            switch=nexus_host)
        if response:
            nexus_type = re.findall(
                "\<[mod:]*desc\>\"*Nexus\s*(\d)\d+\s*[0-9A-Z]+\s*"
                "[cC]hassis\s*\"*\<\/[mod:]*desc\>",
                response)
            if len(nexus_type) > 0:
                LOG.debug("GET call returned Nexus type %d",
                    int(nexus_type[0]))
                return int(nexus_type[0])
        LOG.warning("GET call failed to return Nexus type")
        return -1

    def _extract_line_item_data(self, obj, which, re_str):

        data = obj.re_search_children(which)
        if len(data) == 1:
            data = re.findall(re_str, data[0].text)

        return data[0] if len(data) == 1 else None

    def set_all_vlan_states(self, nexus_host, vlanid_range):
        """Set the VLAN states to active."""
        starttime = time.time()
        LOG.debug("NexusDriver: ")

        snippet = snipp.CMD_VLAN_CREATE_SNIPPET % vlanid_range
        self.capture_and_print_timeshot(
            starttime, "set_all_vlan_states",
            switch=nexus_host)
        self.send_edit_string(nexus_host, None, snippet)

    def get_create_vlan(self, nexus_host, vlanid, vni, conf_str):
        """Returns an XML snippet for create VLAN on a Nexus Switch."""
        LOG.debug("NexusDriver: ")

        starttime = time.time()
        if vni:
            snippet = (snipp.CMD_VLAN_CONF_VNSEGMENT_SNIPPET %
                       (vlanid, vni))
        else:
            snippet = snipp.CMD_VLAN_CREATE_SNIPPET % vlanid

        self.capture_and_print_timeshot(
            starttime, "get_create_vlan",
            switch=nexus_host)

        conf_str += snippet
        return conf_str

    def create_vlan(self, nexus_host, vlanid, vni):
        """Create a VLAN on a Nexus Switch.

        Creates a VLAN given the VLAN ID possibly VxLAN ID.
        """

        LOG.debug("NexusDriver: ")

        starttime = time.time()
        confstr = ''
        confstr = self.get_create_vlan(nexus_host, vlanid, vni, confstr)

        self.send_edit_string(nexus_host, None, confstr,
                              check_to_close_session=False)

        self.capture_and_print_timeshot(
            starttime, "create_vlan_seg",
            switch=nexus_host)

    def delete_vlan(self, nexus_host, vlanid):
        """Delete a VLAN on Nexus Switch given the VLAN ID."""
        confstr = snipp.CMD_NO_VLAN_CONF_SNIPPET % vlanid
        confstr = self.create_xml_snippet(confstr)
        starttime = time.time()
        self._edit_config(nexus_host, target='running', config=confstr,
                          allowed_exc_strs=["None of the VLANs exist"])
        self.capture_and_print_timeshot(
            starttime, "del_vlan",
            switch=nexus_host)

    def _build_intf_confstr(self, snippet, intf_type, interface, vlanid):
        """Build the VLAN config string xml snippet to be used."""
        confstr = snippet % (intf_type, interface, vlanid, intf_type)
        return confstr

    def _get_enable_vlan_on_trunk_int(self, nexus_host, vlanid, intf_type,
                                 interface, is_native, confstr=''):
        """Prepares an XML snippet for VLAN on a trunk interface.

        :param nexus_host: IP address of Nexus switch
        :param vlanid:     Vlanid(s) to add to interface
        :param intf_type:  String which specifies interface type.
                           example: ethernet
        :param interface:  String indicating which interface.
                           example: 1/19
        :param confstr:    last confstr
        :returns           XML snippet
        """
        starttime = time.time()

        snippets = []
        if is_native:
            snippets.append(snipp.CMD_INT_VLAN_NATIVE_SNIPPET)

        snippets.append(snipp.CMD_INT_VLAN_ADD_SNIPPET)

        for snip in snippets:
            confstr += self._build_intf_confstr(
                snippet=snip,
                intf_type=intf_type,
                interface=interface,
                vlanid=vlanid)

        self.capture_and_print_timeshot(
            starttime, "createif",
            switch=nexus_host)

        return confstr

    def disable_vlan_on_trunk_int(self, nexus_host, vlanid, intf_type,
                                  interface, is_native):
        """Disable a VLAN on a trunk interface."""
        starttime = time.time()

        confstr = ''
        if is_native:
            snippet = snipp.CMD_NO_VLAN_INT_NATIVE_SNIPPET
            confstr = (snippet %
                       (intf_type, interface, intf_type))
        snippet = snipp.CMD_NO_VLAN_INT_SNIPPET
        confstr += (snippet %
                   (intf_type, interface, vlanid, intf_type))

        confstr = self.create_xml_snippet(confstr)
        self._edit_config(nexus_host, target='running', config=confstr)
        self.capture_and_print_timeshot(
            starttime, "delif",
            switch=nexus_host)

    def send_edit_string(self, nexus_host, path, confstr,
                         check_to_close_session=True):
        """Sends any XML snippet to Nexus switch."""

        starttime = time.time()
        confstr = self.create_xml_snippet(confstr)
        self._edit_config(nexus_host, target='running',
                          config=confstr,
                          allowed_exc_strs=["VLAN with the same name exists",
                                             "Can't modify state for extended",
                                            "Command is only allowed on VLAN"],
                          check_to_close_session=check_to_close_session)

        self.capture_and_print_timeshot(
            starttime, "send_edit",
            switch=nexus_host)

    def send_enable_vlan_on_trunk_int(self, nexus_host, vlanid, intf_type,
                                      interface, is_native):
        """Gathers and sends an interface trunk XML snippet."""
        confstr = self._get_enable_vlan_on_trunk_int(
                      nexus_host, vlanid,
                      intf_type, interface, is_native)
        self.send_edit_string(nexus_host, None, confstr)

    def create_and_trunk_vlan(self, nexus_host, vlan_id, intf_type,
                              nexus_port, vni, is_native):
        """Create VLAN and trunk it on the specified ports."""
        starttime = time.time()
        self.create_vlan(nexus_host, vlan_id, vni)
        LOG.debug("NexusDriver created VLAN: %s", vlan_id)
        if nexus_port:
            self.send_enable_vlan_on_trunk_int(
                nexus_host, vlan_id,
                intf_type, nexus_port,
                is_native)
        self.capture_and_print_timeshot(
            starttime, "create_all",
            switch=nexus_host)

    def enable_vxlan_feature(self, nexus_host, nve_int_num, src_intf):
        """Enable VXLAN on the switch."""

        starttime = time.time()

        # Configure the "feature" commands and NVE interface
        # (without "member" subcommand configuration).
        # The Nexus 9K will not allow the "interface nve" configuration
        # until the "feature nv overlay" command is issued and installed.
        # To get around the N9K failing on the "interface nve" command
        # send the two XML snippets down separately.
        confstr = self.create_xml_snippet(snipp.CMD_FEATURE_VXLAN_SNIPPET)
        LOG.debug("NexusDriver: ")
        self._edit_config(nexus_host, config=confstr)

        confstr = self.create_xml_snippet((snipp.CMD_INT_NVE_SNIPPET %
                                           (nve_int_num, src_intf)))
        LOG.debug("NexusDriver: ")
        self._edit_config(nexus_host, config=confstr)
        self.capture_and_print_timeshot(
            starttime, "enable_vxlan",
            switch=nexus_host)

    def disable_vxlan_feature(self, nexus_host):
        """Disable VXLAN on the switch."""

        starttime = time.time()

        # Removing the "feature" commands also removes the  NVE interface.
        confstr = self.create_xml_snippet(snipp.CMD_NO_FEATURE_VXLAN_SNIPPET)
        LOG.debug("NexusDriver: ")
        self._edit_config(nexus_host, config=confstr)

        self.capture_and_print_timeshot(
            starttime, "disable_vxlan",
            switch=nexus_host)

    def create_nve_member(self, nexus_host, nve_int_num, vni, mcast_group):
        """Add a member configuration to the NVE interface."""

        starttime = time.time()
        confstr = self.create_xml_snippet((snipp.CMD_INT_NVE_MEMBER_SNIPPET %
                                           (nve_int_num, vni, mcast_group)))
        LOG.debug("NexusDriver: ")
        self._edit_config(nexus_host, config=confstr)

        self.capture_and_print_timeshot(
            starttime, "create_nve",
            switch=nexus_host)

    def delete_nve_member(self, nexus_host, nve_int_num, vni):
        """Delete a member configuration on the NVE interface."""

        starttime = time.time()

        confstr = self.create_xml_snippet((snipp.CMD_INT_NVE_NO_MEMBER_SNIPPET
                                           % (nve_int_num, vni)))
        LOG.debug("NexusDriver: ")
        self._edit_config(nexus_host, config=confstr)

        self.capture_and_print_timeshot(
            starttime, "delete_nve",
            switch=nexus_host)



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function

import argparse
import six

from networking_cisco._i18n import _

from neutronclient.common import extension
from neutronclient.neutron import v2_0 as neutronV20


TEMPLATE = 'hosting_device_template'


def _add_updatable_args(parser):
    parser.add_argument(
        '--name',
        help=_('Name of this hosting device template.'))
    parser.add_argument(
        '--disabled',
        dest='enabled',
        action='store_false',
        help=_('Make the hosting device template disabled.'),
        default=argparse.SUPPRESS)
    parser.add_argument(
        '--service-types',
        help=_('Service types supported by this hosting device template.'))
    parser.add_argument(
        '--service_types',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--image',
        help=_('Glance image used by this hosting device template.'))
    parser.add_argument(
        '--flavor',
        help=_('Nova flavor used by this hosting device template.'))
    parser.add_argument(
        '--credentials-id',
        dest='default_credentials_id',
        help=_('Id of credentials used by default for hosting devices '
               'based on this template.'))
    parser.add_argument(
        '--credentials_id',
        dest='default_credentials_id',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--config-mechanism',
        dest='configuration_mechanism',
        help=_('Method used to configure hosting devices based on this '
               'template.'))
    parser.add_argument(
        '--config_mechanism',
        dest='configuration_mechanism',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--protocol-port',
        help=_('TCP/UDP port used for management of hosting devices based on '
               'this template.'))
    parser.add_argument(
        '--protocol_port',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--booting-time',
        help=_('Typical time to boot hosting devices based on this template.'))
    parser.add_argument(
        '--booting_time',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--tenant-bound',
        help=_('Tenant allowed place service instances in hosting devices '
               'based on this template.'),
        default=argparse.SUPPRESS)
    parser.add_argument(
        '--tenant_bound',
        help=argparse.SUPPRESS,
        default=argparse.SUPPRESS)


def _updatable_args2body(parsed_args, body):
    neutronV20.update_dict(parsed_args, body[TEMPLATE], [
        'name', 'enabled', 'service_types', 'image', 'flavor',
        'default_credentials_id', 'configuration_mechanism', 'protocol_port',
        'booting_time'])
    # handle tenant_bound separately as we want to allow it to be set to None
    if hasattr(parsed_args, 'tenant_bound'):
        if (isinstance(parsed_args.tenant_bound, six.string_types) and
                parsed_args.tenant_bound.lower() == 'none'):
            parsed_args.tenant_bound = None
        body[TEMPLATE].update({'tenant_bound': parsed_args.tenant_bound})


class HostingDeviceTemplate(extension.NeutronClientExtension):
    resource = TEMPLATE
    resource_plural = '%ss' % resource
    object_path = '/dev_mgr/%s' % resource_plural
    resource_path = '/dev_mgr/%s/%%s' % resource_plural
    versions = ['2.0']
    allow_names = True


class HostingDeviceTemplateList(extension.ClientExtensionList,
                                HostingDeviceTemplate):
    """List hosting device templates that belong to a given tenant."""

    shell_command = 'cisco-hosting-device-template-list'
    list_columns = ['id', 'name', 'host_category', 'service_types',
                    'enabled']
    pagination_support = True
    sorting_support = True


class HostingDeviceTemplateShow(extension.ClientExtensionShow,
                                HostingDeviceTemplate):
    """Show information of a given hosting device template."""

    shell_command = 'cisco-hosting-device-template-show'


class HostingDeviceTemplateCreate(extension.ClientExtensionCreate,
                                  HostingDeviceTemplate):
    """Create a hosting device template for a given tenant."""

    shell_command = 'cisco-hosting-device-template-create'

    def add_known_arguments(self, parser):
        _add_updatable_args(parser)
        parser.add_argument(
            '--id',
            help=_('Id for this hosting device template.'))
        parser.add_argument(
            '--slot-capacity',
            help=_('Capacity (in slots) for hosting devices based on this '
                   'template.'))
        parser.add_argument(
            '--slot_capacity',
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--desired-slots-free',
            help=_('Number of slots to keep available in hosting devices '
                   'based on this template.'))
        parser.add_argument(
            '--desired_slots_free',
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--device-driver',
            help=_('Device driver module to use for hosting devices based on '
                   'this template.'))
        parser.add_argument(
            '--device_driver',
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--plugging-driver',
            help=_('Plugging driver module to use for hosting devices based '
                   'on this template.'))
        parser.add_argument(
            '--plugging_driver',
            help=argparse.SUPPRESS)

        parser.add_argument(
            'name', metavar='NAME',
            help=_('Name of this hosting device template.'))
        parser.add_argument(
            'host_category', metavar='HOST_CATEGORY',
            help=_('Host category for this hosting device template. One of '
                   'VM, Hardware, or, Network_Node.'))

    def args2body(self, parsed_args):
        body = {self.resource: {'name': parsed_args.name,
                                'host_category': parsed_args.host_category}}
        _updatable_args2body(parsed_args, body)
        neutronV20.update_dict(parsed_args, body[self.resource],
                               ['id', 'tenant_id', 'slot_capacity',
                                'desired_slots_free', 'device_driver',
                                'plugging_driver'])
        return body


class HostingDeviceTemplateDelete(extension.ClientExtensionDelete,
                                  HostingDeviceTemplate):
    """Delete a given hosting device template."""

    shell_command = 'cisco-hosting-device-template-delete'


class HostingDeviceTemplateUpdate(extension.ClientExtensionUpdate,
                                  HostingDeviceTemplate):

    """Update hosting device template's information."""

    shell_command = 'cisco-hosting-device-template-update'

    def add_known_arguments(self, parser):
        # adding enable here so that it is available for update only as it is
        # True by default and not meaningful in the create operation
        parser.add_argument(
            '--enabled',
            dest='enabled',
            action='store_true',
            help=_('Make the hosting device template enabled.'),
            default=argparse.SUPPRESS)
        _add_updatable_args(parser)

    def args2body(self, parsed_args):
        body = {self.resource: {}}
        _updatable_args2body(parsed_args, body)
        return body



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock
from oslo_utils import uuidutils
from webob import exc

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import utils
from networking_cisco.plugins.cisco.extensions import routertype

from neutron.tests import base
from neutron.tests.unit.api.v2 import test_base
from neutron.tests.unit.extensions import base as test_extensions_base


_uuid = uuidutils.generate_uuid
_get_path = test_base._get_path
NEUTRON_VERSION = bc.NEUTRON_VERSION
NEUTRON_NEWTON_VERSION = bc.NEUTRON_NEWTON_VERSION


class RouterTypeTestCase(test_extensions_base.ExtensionTestCase):
    fmt = 'json'

    def setUp(self):
        super(RouterTypeTestCase, self).setUp()
        if NEUTRON_VERSION.version[0] > NEUTRON_NEWTON_VERSION.version[0]:
            plugin = ('networking_cisco.plugins.cisco.service_plugins.'
                      'cisco_router_plugin.CiscoRouterPlugin')
            service_type = bc.constants.L3
        else:
            # NOTE(bobmel): The routertype extension is for the router service.
            # We therefore add 'router' to supported extensions of the core
            # plugin used in these tests. That way, NeutronManager will return
            # that plugin as the l3 router service plugin.
            plugin = ('networking_cisco.plugins.cisco.extensions.routertype.'
                      'RoutertypePluginBase')
            service_type = None
        self._setUpExtension(
            plugin, service_type,
            routertype.RESOURCE_ATTRIBUTE_MAP, routertype.Routertype, '',
            supported_extension_aliases=['router',
                                         routertype.ROUTERTYPE_ALIAS])

    def test_create_routertype(self):
        dummy_driver = ('networking_cisco.plugins.cisco.l3.schedulers.'
                        'noop_l3_router_hosting_device_scheduler.'
                        'NoopL3RouterHostingDeviceScheduler')
        rt_id = _uuid()
        tenant_id = _uuid()
        data = {'routertype': {
            'id': None,
            'tenant_id': tenant_id,
            'name': 'Fancy router type 1',
            'description': 'Lightning fast router type',
            'template_id': _uuid(),
            'ha_enabled_by_default': False,
            'shared': True,
            'slot_need': 200,
            'scheduler': dummy_driver,
            'driver': dummy_driver,
            'cfg_agent_service_helper': dummy_driver,
            'cfg_agent_driver': dummy_driver}}

        if bc.NEUTRON_VERSION >= bc.NEUTRON_NEWTON_VERSION:
            data['routertype']['project_id'] = tenant_id

        return_value = copy.copy(data['routertype'])
        return_value.update({'id': rt_id})

        instance = self.plugin.return_value
        instance.create_routertype.return_value = return_value
        res = self.api.post(_get_path('routertypes', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        instance.create_routertype.assert_called_with(mock.ANY,
                                                      routertype=data)
        self.assertEqual(res.status_int, exc.HTTPCreated.code)
        res = self.deserialize(res)
        self.assertIn('routertype', res)
        self.assertEqual(res['routertype'], return_value)

    def test_routertype_list(self):
        rt_id = _uuid()
        return_value = [{'tenant_id': _uuid(),
                         'id': rt_id}]

        instance = self.plugin.return_value
        instance.get_routertypes.return_value = return_value

        res = self.api.get(_get_path('routertypes', fmt=self.fmt))

        instance.get_routertypes.assert_called_with(mock.ANY, fields=mock.ANY,
                                                    filters=mock.ANY)
        self.assertEqual(res.status_int, exc.HTTPOk.code)

    def test_routertype_get(self):
        rt_id = _uuid()
        return_value = {'tenant_id': _uuid(),
                        'id': rt_id}

        instance = self.plugin.return_value
        instance.get_routertype.return_value = return_value

        res = self.api.get(_get_path('routertypes',
                                     id=rt_id, fmt=self.fmt))

        instance.get_routertype.assert_called_with(mock.ANY, rt_id,
                                                   fields=mock.ANY)
        self.assertEqual(res.status_int, exc.HTTPOk.code)
        res = self.deserialize(res)
        self.assertIn('routertype', res)
        self.assertEqual(res['routertype'], return_value)

    def test_routertype_update(self):
        rt_id = _uuid()
        update_data = {'routertype': {'name': 'Even fancier router type'}}
        return_value = {'tenant_id': _uuid(), 'id': rt_id}

        instance = self.plugin.return_value
        instance.update_routertype.return_value = return_value

        res = self.api.put(_get_path('routertypes', id=rt_id, fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_routertype.assert_called_with(
            mock.ANY, rt_id, routertype=update_data)
        self.assertEqual(res.status_int, exc.HTTPOk.code)
        res = self.deserialize(res)
        self.assertIn('routertype', res)
        self.assertEqual(res['routertype'], return_value)

    def test_routertype_delete(self):
        self._test_entity_delete('routertype')


class TestRouterTypeAttributeValidators(base.BaseTestCase):

    def test_convert_validate_driver(self):
        drv = ('networking_cisco.plugins.cisco.cfg_agent.device_drivers.'
               'dummy_driver.DummyRoutingDriver')
        res = utils.convert_validate_driver_class(drv)
        self.assertEqual(res, drv)

        self.assertRaises(utils.DriverNotFound,
                          utils.convert_validate_driver_class,
                          'this.is.not.a.driver')



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import binascii
import eventlet
import os
import signal
import socket
import struct

from neutron.common import config
from oslo_config import cfg
from oslo_log import log as logging

from networking_cisco._i18n import _
from networking_cisco.plugins.cisco.cpnr.cpnr_client import UnexpectedError
from networking_cisco.plugins.cisco.cpnr import debug_stats
from networking_cisco.plugins.cisco.cpnr import netns

LOG = logging.getLogger(__name__)

DEBUG_STATS_MIN_WRITE_INTERVAL = 30
MONITOR_INTERVAL = 1
RECV_BUFFER_SIZE = 4096
NS_RELAY_PENDING = 'NS_RELAY_PENDING'
NS_RELAY_RUNNING = 'NS_RELAY_RUNNING'
NS_RELAY_DELETING = 'NS_RELAY_DELETING'
DHCP_CLIENT_PORT = 68
DHCP_SERVER_PORT = 67
RLIMIT_NOFILE_LIMIT = 16384

# Relay agent information options.
LINK_SELECTION = 5
SERVER_IDENTIFIER_OVERRIDE = 11
RELAY_AGENT_INFO = 82
VSS = 151
VSS_CONTROL = 152
END = 255

opt_cpnr_group = cfg.OptGroup(
    name='cisco_pnr', title='Cisco Prime Network Registrar configuration')

OPTS = [
    cfg.StrOpt('http_server',
               default="localhost:8080",
               help=_('External HTTP server, should conform to '
                      '<server_name:port> format.')),
    cfg.StrOpt('admin_email',
               default='test@example.com',
               help=_('Email address of admin for internal DNS domain.')),
    cfg.StrOpt('external_interface',
               default='lo',
               help=_('Interface for communicating with DHCP/DNS server.')),
    cfg.StrOpt('dhcp_server_addr',
               default='127.0.0.1',
               help=_('DHCP server IP address.')),
    cfg.IntOpt('dhcp_server_port',
               default=67,
               help=_('DHCP server UDP port number.')),
    cfg.BoolOpt('enable_dhcp_stats',
                default=False,
                help=_('Enable DHCP stats.')),
    cfg.IntOpt('dhcp_stats_interval',
               default=60,
               help=_('DHCP stats polling interval.'))
]


class DhcpRelayAgent(object):
    """Relay DHCP packets between neutron networks and external DHCP server.

    Receives broadcast and unicast DHCP requests via sockets which are opened
    in each neutron dhcp network namespace.  Additional DHCP options are
    appended to the request to indicate from which network the request
    originated. Requests are then forwarded to the configured DHCP server
    address.

    Receives unicast DHCP responses from the DHCP server via socket opened in
    the global network namespace.  Additional options are stripped from the
    response. The response is then forwarded to the originating network.
    """

    def __init__(self):
        self.conf = cfg.CONF
        self.conf.register_group(opt_cpnr_group)
        self.conf.register_opts(OPTS, opt_cpnr_group)
        LOG.debug('(cisco_pnr) http_server: {}'.format(
            self.conf.cisco_pnr.http_server))
        LOG.debug('(cisco_pnr) external_interface: {}'.format(
            self.conf.cisco_pnr.external_interface))
        LOG.debug('(cisco_pnr) dhcp_server_addr: {}'.format(
            self.conf.cisco_pnr.dhcp_server_addr))
        LOG.debug('(cisco_pnr) dhcp_server_port: {}'.format(
            self.conf.cisco_pnr.dhcp_server_port))
        self.ns_states = {}
        self.int_sockets_by_vpn = {}
        self.ext_sock = None
        self.ext_addr = ""
        self.ns_lock = eventlet.semaphore.Semaphore()
        self.int_sock_retries = 0
        self.debug_stats = debug_stats.DebugStats('dhcp')
        self.kill_now = False

    def serve(self):
        self.greenpool = eventlet.GreenPool(3)
        self.greenpool.spawn_n(self._server_network_relay)
        self.greenpool.spawn_n(self._namespace_monitor)
        if self.conf.cisco_pnr.enable_dhcp_stats:
            self.greenpool.spawn_n(self._write_debug_stats)
        self.greenpool.waitall()

    def _signal_handler(self, signum, frame):
        LOG.debug('Recieved the signal %s', signum)
        self.kill_now = True

    def _namespace_monitor(self):

        while True:
            if self.kill_now:
                break
            eventlet.sleep(MONITOR_INTERVAL)

            # Get list of network namespaces on system
            try:
                curr_ns = set(netns.nslist())
            except Exception:
                LOG.error('Failed to get current namespace set')
                continue

            # For each unknown namespace, start a relay thread
            for ns in curr_ns:
                if not ns.startswith("qdhcp") or ns in self.ns_states:
                    continue
                self.ns_states[ns] = NS_RELAY_PENDING
                eventlet.spawn_n(self._client_network_relay, ns)

            # Set state to DELETING for any unknown namespaces
            for ns in self.ns_states:
                if ns in curr_ns:
                    continue
                self.ns_states[ns] = NS_RELAY_DELETING
        LOG.debug('Namespace Monitor exiting')

    def _server_network_relay(self):

        # Open a socket in the global namespace for DHCP
        try:
            self.ext_sock, self.ext_addr = self._open_dhcp_ext_socket()
        except Exception:
            LOG.exception("Failed to open dhcp external socket in "
                          "global ns")
            return
        recvbuf = bytearray(RECV_BUFFER_SIZE)

        # Forward DHCP responses from external to internal networks
        while True:
            if self.kill_now:
                break
            try:
                self.ext_sock.settimeout(1)
                size = self.ext_sock.recv_into(recvbuf)
                pkt = DhcpPacket.parse(recvbuf)
                vpnid = pkt.get_relay_option(151)
                ciaddr = pkt.get_ciaddr()
                if vpnid not in self.int_sockets_by_vpn:
                    continue
                int_sock = self.int_sockets_by_vpn[vpnid]
                self.debug_stats.increment_pkts_from_server(vpnid)
                if ciaddr == "0.0.0.0":
                    ciaddr = "255.255.255.255"
                LOG.debug('Forwarding DHCP response for vpn %s', vpnid)
                int_sock.sendto(recvbuf[:size], (ciaddr, DHCP_CLIENT_PORT))
                self.debug_stats.increment_pkts_to_client(vpnid)
            except socket.timeout:
                pass
            except Exception:
                LOG.exception('Failed to forward dhcp response')
        LOG.debug('Server network relay exiting')

    def _client_network_relay(self, namespace):

        # Open a socket in the DHCP network namespace
        try:
            with self.ns_lock, netns.Namespace(namespace):
                recv_sock, send_sock, int_addr = self._open_dhcp_int_socket()
        except Exception:
            self.int_sock_retries += 1
            if self.int_sock_retries >= 2:
                LOG.exception('Failed to open dhcp server socket in %s',
                              namespace)
                self.int_sock_retries = 0
            del self.ns_states[namespace]
            return
        self.int_sock_retries = 0
        self.ns_states[namespace] = NS_RELAY_RUNNING
        vpnid = self._convert_ns_to_vpnid(namespace)
        self.debug_stats.add_network_stats(vpnid)
        self.int_sockets_by_vpn[vpnid] = send_sock
        recvbuf = bytearray(RECV_BUFFER_SIZE)
        LOG.debug('Opened dhcp server socket on ns:%s, addr:%s, vpn:%s',
                  namespace, int_addr, vpnid)
        # Forward DHCP requests from internal to external networks
        while self.ns_states[namespace] != NS_RELAY_DELETING:
            if self.kill_now:
                break
            try:
                recv_sock.settimeout(1)
                recv_sock.recv_into(recvbuf)
                pkt = DhcpPacket.parse(recvbuf)
                options = [(5, int_addr),
                           (11, int_addr),
                           (151, vpnid),
                           (152, '')]
                for option in options:
                    pkt.set_relay_option(*option)
                pkt.set_giaddr(self.ext_addr)
                self.debug_stats.increment_pkts_from_client(vpnid)
                LOG.debug('Forwarding DHCP request for vpn %s', vpnid)
                self.ext_sock.send(pkt.data())
                self.debug_stats.increment_pkts_to_server(vpnid)
            except socket.timeout:
                pass
            except Exception:
                LOG.exception('Failed to forward dhcp to server from %s',
                              namespace)

        # Cleanup socket and internal state
        try:
            del self.ns_states[namespace]
            del self.int_sockets_by_vpn[vpnid]
            self.debug_stats.del_network_stats(vpnid)
            recv_sock.close()
            send_sock.close()
        except Exception:
            LOG.warning('Failed to cleanup relay for %s', namespace)
        LOG.debug('Client network relay exiting')

    def _open_dhcp_ext_socket(self):

        # find configured external interface ip address
        for ifname, addr, mask in netns.iflist():
            if ifname == self.conf.cisco_pnr.external_interface:
                break
        else:
            raise UnexpectedError(msg='Failed to find external intf '
                                      'matching config')

        # open, bind, and connect UDP socket
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.bind((addr, DHCP_SERVER_PORT))
        s.connect((self.conf.cisco_pnr.dhcp_server_addr,
                   self.conf.cisco_pnr.dhcp_server_port))
        return s, addr

    def _open_dhcp_int_socket(self):

        # list interfaces, fail if not exactly one
        interfaces = netns.iflist(ignore=("lo",))
        if not interfaces:
            raise UnexpectedError(msg="failed to find single interface "
                                      "in dhcp ns")
        addr = interfaces[0][1]

        # open socket for receiving DHCP requests on internal net
        recv_s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        recv_s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        recv_s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        recv_s.bind(("0.0.0.0", DHCP_SERVER_PORT))

        # open socket for sending DHCP responses on internal net
        send_s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        send_s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        send_s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        send_s.bind((addr, DHCP_SERVER_PORT))

        return recv_s, send_s, addr

    def _convert_ns_to_vpnid(self, ns):
        return ns.replace('-', '')[-14:]

    def _write_debug_stats(self):
        polling_interval = max(DEBUG_STATS_MIN_WRITE_INTERVAL,
                               self.conf.cisco_pnr.dhcp_stats_interval)
        while True:
            eventlet.sleep(polling_interval)
            self.debug_stats.write_stats_to_file()


class DhcpPacket(object):

    def __init__(self):
        self.buf = ''
        self.ciaddr = ''
        self.giaddr = ''
        self.relay_options = {}

    @classmethod
    def parse(cls, buf):
        """Parse DHCP Packet.

        1. To get client IP Address(ciaddr).
        2. To get relaying gateway IP Address(giaddr).
        3. To get DHCP Relay Agent Information Option Suboption
            such as Link Selection, VSS, Server Identifier override.
        """

        pkt = DhcpPacket()
        (pkt.ciaddr,) = cls.struct('4s').unpack_from(buf, 12)
        (pkt.giaddr,) = cls.struct('4s').unpack_from(buf, 24)
        cls.struct('4s').pack_into(buf, 24, b'')
        pos = 240
        while pos < len(buf):
            (opttag,) = cls.struct('B').unpack_from(buf, pos)
            if opttag == 0:
                pos += 1
                continue
            if opttag == END:
                pkt.end = pos
                break
            (optlen,) = cls.struct('B').unpack_from(buf, pos + 1)
            startpos = pos
            pos += 2
            if opttag != RELAY_AGENT_INFO:
                pos += optlen
                continue
            optend = pos + optlen
            while pos < optend:
                (subopttag, suboptlen) = cls.struct('BB').unpack_from(buf, pos)
                fmt = '%is' % (suboptlen,)
                (val,) = cls.struct(fmt).unpack_from(buf, pos + 2)
                pkt.relay_options[subopttag] = val
                pos += suboptlen + 2
            cls.struct('%is' % (optlen + 2)).pack_into(buf, startpos, b'')
        pkt.buf = buf
        return pkt

    def get_relay_option(self, code):
        value = self.relay_options[code]
        if code in (LINK_SELECTION, SERVER_IDENTIFIER_OVERRIDE):
            value = socket.inet_ntoa(value)
        elif code == VSS:
            value = binascii.hexlify(value[1:])
        return value

    def set_relay_option(self, code, value):
        if code == LINK_SELECTION or code == SERVER_IDENTIFIER_OVERRIDE:
            value = socket.inet_aton(value)
        elif code == VSS:
            value = binascii.unhexlify("01" + value)
        self.relay_options[code] = value

    def get_ciaddr(self):
        return socket.inet_ntoa(self.ciaddr)

    def set_giaddr(self, addr):
        self.giaddr = socket.inet_aton(addr)

    def data(self):
        self.struct('4s').pack_into(self.buf, 12, self.ciaddr)
        self.struct('4s').pack_into(self.buf, 24, self.giaddr)
        opttag = 82
        optlen = 0
        for val in self.relay_options.values():
            optlen += len(val) + 2
        self.struct('BB').pack_into(self.buf, self.end, opttag, optlen)
        self.end += 2
        for code, val in self.relay_options.items():
            fmt = 'BB%is' % (len(val),)
            self.struct(fmt).pack_into(self.buf, self.end, code, len(val), val)
            self.end += len(val) + 2
        self.struct('B').pack_into(self.buf, self.end, 255)
        return self.buf[:self.end + 1]

    structcache = {}

    @classmethod
    def struct(cls, fmt):
        return cls.structcache.setdefault(fmt, struct.Struct(fmt))


def main():
    try:
        netns.increase_ulimit(RLIMIT_NOFILE_LIMIT)
    except Exception:
        LOG.error('Failed to increase ulimit for DHCP relay')
    eventlet.monkey_patch()
    config.setup_logging()
    if os.getuid() != 0:
        LOG.error('Must run dhcp relay as root')
        return
    relay = DhcpRelayAgent()
    signal.signal(signal.SIGINT, relay._signal_handler)
    signal.signal(signal.SIGTERM, relay._signal_handler)
    relay.serve()

if __name__ == "__main__":
    main()



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging
import oslo_messaging

from neutron.common import rpc as n_rpc

from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.extensions import ciscocfgagentscheduler

LOG = logging.getLogger(__name__)


CFGAGENT_SCHED = ciscocfgagentscheduler.CFG_AGENT_SCHEDULER_ALIAS


class DeviceMgrCfgAgentNotifyAPI(object):
    """API for Device manager service plugin to notify Cisco cfg agent."""

    def __init__(self, devmgr_plugin, topic=c_constants.CFG_AGENT):
        self._dmplugin = devmgr_plugin
        target = oslo_messaging.Target(topic=topic, version='1.1')
        self.client = n_rpc.get_client(target)

    def _host_notification(self, context, method, payload, host):
        """Notify the cfg agent that is handling the hosting device."""
        LOG.debug('Notify Cisco cfg agent at %(host)s the message '
                  '%(method)s', {'host': host, 'method': method})
        cctxt = self.client.prepare(server=host)
        cctxt.cast(context, method, payload=payload)

    def _agent_notification(self, context, method, hosting_devices, operation):
        """Notify individual Cisco cfg agents."""
        admin_context = context.is_admin and context or context.elevated()
        for hosting_device in hosting_devices:
            agents = self._dmplugin.get_cfg_agents_for_hosting_devices(
                admin_context, hosting_device['id'], admin_state_up=True,
                schedule=True)
            for agent in agents:
                LOG.debug('Notify %(agent_type)s at %(topic)s.%(host)s the '
                          'message %(method)s',
                          {'agent_type': agent.agent_type,
                           'topic': agent.topic,
                           'host': agent.host,
                           'method': method})
                cctxt = self.client.prepare(server=agent.host)
                cctxt.cast(context, method)

    def agent_updated(self, context, admin_state_up, host):
        """Updates cfg agent on <host> to enable or disable it."""
        self._host_notification(context, 'agent_updated',
                                {'admin_state_up': admin_state_up}, host)

    def hosting_devices_unassigned_from_cfg_agent(self, context, ids, host):
        """Notify cfg agent to no longer handle some hosting devices.

        This notification relieves the cfg agent in <host> of responsibility
        to monitor and configure hosting devices with id specified in <ids>.
        """
        self._host_notification(context,
                                'hosting_devices_unassigned_from_cfg_agent',
                                {'hosting_device_ids': ids}, host)

    def hosting_devices_assigned_to_cfg_agent(self, context, ids, host):
        """Notify cfg agent to now handle some hosting devices.

        This notification relieves the cfg agent in <host> of responsibility
        to monitor and configure hosting devices with id specified in <ids>.
        """
        self._host_notification(context,
                                'hosting_devices_assigned_to_cfg_agent',
                                {'hosting_device_ids': ids}, host)

    def hosting_devices_removed(self, context, hosting_data, deconfigure,
                                host):
        """Notify cfg agent that some hosting devices have been removed.

        This notification informs the cfg agent in <host> that the
        hosting devices in the <hosting_data> dictionary have been removed
        from the hosting device pool. The <hosting_data> dictionary also
        contains the ids of the affected logical resources for each hosting
        devices::

             {'hd_id1': {'routers': [id1, id2, ...],
                         'fw': [id1, ...],
                         ...},
              'hd_id2': {'routers': [id3, id4, ...]},
                         'fw': [id1, ...],
                         ...},
              ...}

        The <deconfigure> argument is True if any configurations for the
        logical resources should be removed from the hosting devices
        """
        if hosting_data:
            self._host_notification(context, 'hosting_devices_removed',
                                    {'hosting_data': hosting_data,
                                     'deconfigure': deconfigure}, host)

    # version 1,1
    def get_hosting_device_configuration(self, context, id):
        """Fetch configuration of hosting device with id.

        The configuration agent should respond with the running config of
        the hosting device.
        """
        admin_context = context.is_admin and context or context.elevated()
        agents = self._dmplugin.get_cfg_agents_for_hosting_devices(
            admin_context, [id], admin_state_up=True, schedule=True)
        if agents:
            cctxt = self.client.prepare(server=agents[0].host)
            return cctxt.call(context, 'get_hosting_device_configuration',
                              payload={'hosting_device_id': id})



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import wraps
import time

from oslo_log import log as logging
from oslo_utils import importutils

from neutron_lib import exceptions as nexception

from networking_cisco._i18n import _

LOG = logging.getLogger(__name__)


class DriverNotFound(nexception.NotFound):
    message = _("Driver %(driver)s does not exist")


def retry(ExceptionToCheck, tries=4, delay=3, backoff=2):
    """Retry calling the decorated function using an exponential backoff.

    Reference: http://www.saltycrane.com/blog/2009/11/trying-out-retry
    -decorator-python/

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :param tries: number of times to try (not retry) before giving up
    :param delay: initial delay between retries in seconds
    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    """
    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    LOG.debug("%(err_mess)s. Retry calling function "
                              "'%(f_name)s' in %(delta)d seconds.",
                              {'err_mess': str(e), 'f_name': f.__name__,
                               'delta': mdelay})
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            LOG.debug("Last retry calling function '%s'.", f.__name__)
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry


def convert_validate_driver_class(driver_class_name):
    # Verify that import_obj is a loadable class
    if driver_class_name is None or driver_class_name == '':
        return driver_class_name
    else:
        try:
            if importutils.import_class(driver_class_name):
                return driver_class_name
        except ImportError as e:
            LOG.error('Failed to verify driver module %(name)s: %(err)s',
                      {'name': driver_class_name, 'err': e})
    raise DriverNotFound(driver=driver_class_name)



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


from __future__ import print_function

import ConfigParser
import optparse
import os
import platform
import re
import shlex
import subprocess as subp
import sys


NEUTRON = 'neutron'
KEYSTONE = 'keystone'

conf_file_list = [
    'keystone.conf',
    'neutron.conf'
]
default_path = '/etc/neutron,/etc/keystone'
dfa_cfg_file = '/etc/saf/enabler_conf.ini'

dfa_neutron_option_list = [
    {'section': 'DEFAULT',
     'option': 'rpc_backend',
     'value': 'rabbit',
     'is_list': False},
    {'section': 'DEFAULT',
     'option': 'notification_driver',
     'value': 'messaging',
     'is_list': False},
    {'section': 'DEFAULT',
     'option': 'notification_topics',
     'value': 'cisco_dfa_neutron_notify',
     'is_list': True},
]
dfa_keystone_option_list = [
    {'section': 'DEFAULT',
     'option': 'rpc_backend',
     'value': 'rabbit',
     'is_list': False},
    {'section': 'DEFAULT',
     'option': 'notification_driver',
     'value': 'messaging',
     'is_list': False},
    {'section': 'DEFAULT',
     'option': 'notification_topics',
     'value': 'cisco_dfa_keystone_notify',
     'is_list': True},
]

service_options = {
    'neutron': dfa_neutron_option_list,
    'keystone': dfa_keystone_option_list,
}


dist_data = {
    'ubuntu': {'init_dir': '/etc/init/',
               'server_conf': 'fabric-enabler-server.conf',
               'agent_conf': 'fabric-enabler-agent.conf'},
    'centos': {'init_dir': '/usr/lib/systemd/system/',
               'server_conf': 'fabric-enabler-server.service',
               'agent_conf': 'fabric-enabler-agent.service'},
    'redhat': {'init_dir': '/usr/lib/systemd/system/',
               'server_conf': 'fabric-enabler-server.service',
               'agent_conf': 'fabric-enabler-agent.service'},
}


def get_cmd_output(cmd):
    final_cmd = shlex.split(cmd)
    try:
        output = subp.check_output(final_cmd)
    except subp.CalledProcessError as exc:
        print("Error running %s: error: %s, output: %s" % (
              cmd, exc.returncode, exc.output))
        sys.exit(0)
    except Exception as exc:
        print("Exception %s running command %s" % (cmd, exc))
        sys.exit(0)

    return output


def get_mysql_credentials(cfg_file):
    """Get the credentials and database name from options in config file."""

    try:
        parser = ConfigParser.ConfigParser()
        cfg_fp = open(cfg_file)
        parser.readfp(cfg_fp)
        cfg_fp.close()
    except ConfigParser.NoOptionError:
        cfg_fp.close()
        print('Failed to find mysql connections credentials.')
        sys.exit(1)
    except IOError:
        print('ERROR: Cannot open %s.', cfg_file)
        sys.exit(1)

    value = parser.get('dfa_mysql', 'connection')

    try:
        # Find location of pattern in connection parameter as shown below:
        # http://username:password@host/databasename?characterset=encoding'
        sobj = re.search(r"(://).*(@).*(/).*(\?)", value)

        # The list parameter contains:
        # indices[0], is the index of '://'
        # indices[1], is the index of '@'
        # indices[2], is the index of '/'
        # indices[3], is the index of '?'
        indices = [sobj.start(1), sobj.start(2), sobj.start(3), sobj.start(4)]

        # Get the credentials
        cred = value[indices[0] + 3:indices[1]].split(':')

        # Get the host name
        host = value[indices[1] + 1:indices[2]]

        # Get the database name
        db_name = value[indices[2] + 1:indices[3]]

        # Get the character encoding
        charset = value[indices[3] + 1:].split('=')[1]

        return cred[0], cred[1], host, db_name, charset
    except (ValueError, IndexError, AttributeError):
        print('Failed to find mysql connections credentials.')
        sys.exit(1)


def modify_conf(cfgfile, service_name, outfn):
    """Modify config file neutron and keystone to include enabler options."""

    if not cfgfile or not outfn:
        print('ERROR: There is no config file.')
        sys.exit(0)

    options = service_options[service_name]
    with open(cfgfile, 'r') as cf:
        lines = cf.readlines()

    for opt in options:
        op = opt.get('option')
        res = [line for line in lines if line.startswith(op)]
        if len(res) > 1:
            print('ERROR: There are more than one %s option.' % res)
            sys.exit(0)
        if res:
            (op, sep, val) = (res[0].strip('\n').replace(' ', '').
                              partition('='))
            new_val = None
            if opt.get('is_list'):
                # Value for this option can contain list of values.
                # Append the value if it does not exist.
                if not any(opt.get('value') == value for value in
                           val.split(',')):
                    new_val = ','.join((val, opt.get('value')))
            else:
                if val != opt.get('value'):
                    new_val = opt.get('value')
            if new_val:
                opt_idx = lines.index(res[0])
                # The setting is different, replace it with new one.
                lines.pop(opt_idx)
                lines.insert(opt_idx, '='.join((opt.get('option'),
                             new_val + '\n')))
        else:
            # Option does not exist. Add the option.
            try:
                sec_idx = lines.index('[' + opt.get('section') + ']\n')
                lines.insert(sec_idx + 1, '='.join(
                    (opt.get('option'), opt.get('value') + '\n')))
            except ValueError:
                print('Invalid %s section name.' % opt.get('section'))
                sys.exit(0)

    with open(outfn, 'w') as fwp:
        all_lines = ''
        for line in lines:
            all_lines += line

        fwp.write(all_lines)


def prepare_db():

    (user, password, host, db, charset) = get_mysql_credentials(dfa_cfg_file)

    # Modify max_connections, if it is not 2000
    get_var_cmd = ('mysql -e '
                   '"show variables like \'max_connections\';"')
    out = get_cmd_output(get_var_cmd)
    try:
        val = int(out.split('\n')[1].split('\t')[1])
    except Exception:
        print('Invalid value: Cannot get max_connections from DB.')
        sys.exit(0)

    if val < 2000:
        # Set max_connections to 2000 if it is not.
        set_conn_cmd = ('mysql -e '
                        '"set global max_connections = 2000;"')
        out = get_cmd_output(set_conn_cmd)
        print(out)

    # Create database if it not existed.
    create_cmd = ('mysql '
                  '-e "CREATE DATABASE IF NOT EXISTS %(db)s '
                  'CHARACTER SET %(charset)s;"' % (
                      {'user': user, 'password': password, 'host': host,
                       'db': db, 'charset': charset}))
    out = get_cmd_output(create_cmd)
    print(out)

    # Create user for enabler if it does not exist.
    check_user_cmd = ('mysql -e '
                      '"SELECT EXISTS(SELECT DISTINCT user FROM mysql.user'
                      ' WHERE user=\'%s\' AND host=\'%s\')as user;"' % (
                          user, host))
    out = get_cmd_output(check_user_cmd)
    if int(out.split()[1]) == 0:
        # User does not exist. Create new one.
        create_user_cmd = ('mysql -e '
                           '"CREATE USER \'%(user)s\'@\'%(host)s\''
                           'IDENTIFIED BY \'%(pwd)s\';"' % {
                               'user': user, 'host': host, 'pwd': password})
        out = get_cmd_output(create_user_cmd)
        if 'ERROR' in out:
            print('Failed to create %(user)s in MySQL.\n%(reason)s') % (
                {'user': user, 'reason': out})
            sys.exit(0)

    # Grant permission to the user.
    grant_perm_cmd = ("mysql -e "
                      "\"GRANT  ALL PRIVILEGES ON *.* TO "
                      "'%(user)s'@'%(host)s';\"") % (
                          {'user': user, 'host': host})
    out = get_cmd_output(grant_perm_cmd)
    if 'ERROR' in out:
        print('Failed to grant permission to %(user)s.\n%(reason)s') % (
            {'user': user, 'reason': out})
        sys.exit(0)


def find_conf_and_modify(os_path, root_helper):

    # Search for the config files in the path
    for path in os_path.split(','):
        for p, d, f in os.walk(path):
            for fn in f:
                if fn in conf_file_list:
                    fname = os.path.realpath(os.path.join(p, fn))
                    service_name = NEUTRON if NEUTRON in fname else (
                        KEYSTONE if KEYSTONE in fname else None)
                    modify_conf(fname, service_name, fname + '.modified')

                    # Keep the existing in .orig and copy the .modified
                    # to the exisiting one.
                    cmd = root_helper + 'cp %s %s.orig' % (fname, fname)
                    print(cmd)
                    get_cmd_output(cmd)
                    cmd = root_helper + 'cp %s.modified %s' % (fname, fname)
                    print(cmd)
                    get_cmd_output(cmd)


def copy_init_conf_files(node, root_helper):

    # Copy fabric-enabler-server and fabric-enabler-agent
    # to init directory based on Linux distribution.
    path = (os.path.dirname(os.path.dirname(os.path.abspath(__file__))) +
            '/etc/saf/init/')
    dist = platform.dist()[0].lower()
    if dist not in dist_data:
        print('This %s Linux distribution is not supported.') % dist
        sys.exit(1)

    init_dir = dist_data[dist].get('init_dir')
    conf_fn = None
    if node == 'control':
        conf_fn = dist_data[dist].get('server_conf')
    if node == 'compute':
        conf_fn = dist_data[dist].get('agent_conf')

    if conf_fn is not None:
        cmd = root_helper + 'cp %s %s' % ((path + conf_fn), init_dir)
        print(cmd)
        get_cmd_output(cmd)
        if dist == 'centos' or dist == 'redhat':
            cmd3 = root_helper + 'systemctl enable %s' % conf_fn
            print(cmd3)
            get_cmd_output(cmd3)


usage = ('\n'
         'python dfa_prepare_setup.py --dir-path filepath1[,filepath2,...]'
         '--node-function [control | compute]\n')

if __name__ == '__main__':

    root_helper = ''
    if os.geteuid() != 0:
        # This is not root
        root_helper = 'sudo '
    parser = optparse.OptionParser(usage=usage)

    parser.add_option('--dir-path',
                      type='string', dest='dir_path', default=default_path,
                      help='Path to neutron.conf and keystone.conf files')
    parser.add_option('--node-function',
                      type='string', dest='node_function', default='control',
                      help='Choose the node runs as controller or compute.')
    (options, args) = parser.parse_args()

    node = options.node_function.lower()
    if node == 'control':
        find_conf_and_modify(options.dir_path.lower(), root_helper)
        prepare_db()

    copy_init_conf_files(node, root_helper)
